from collections import defaultdict
from functools import wraps
from typing import Any, Iterator

from addict import Dict

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import Integer, String, Structure, UnsignedFloat

__all__ = ["ComponentSystem", "Component", "Species"]


class Species(Structure):
    """
    Species class.

    Represent a species in a chemical system.

    Attributes
    ----------
    name : str
        The name of the species.
    charge : int, optional
        The charge of the species. Default is 0.
    molecular_weight : float
        The molecular weight of the species.
    density : float
        Density of the species.
    """

    name: String = String()
    charge: Integer = Integer(default=0)
    molecular_weight: UnsignedFloat = UnsignedFloat()
    density: UnsignedFloat = UnsignedFloat()

    def __str__(self) -> str:
        """str: String representation of the component."""
        return self.name


class Component(Structure):
    """
    Information about single component.

    A component can contain subspecies (e.g. differently charged variants).

    Attributes
    ----------
    name : str | None
        Name of the component.
    species : list[Species]
        List of Subspecies.
    n_species : int
        Number of Subspecies.
    label : list[str]
        Name of component (including species).
    charge : int | list[int | None]
        Charge of component (including species).
    molecular_weight : float | list[float | None]
        Molecular weight of component (including species).
    density : float | list[float | None]
        Density of component (including species).

    See Also
    --------
    Species
    ComponentSystem
    """

    name: String = String()

    def __init__(
        self,
        name: str | None = None,
        species: str | list[str | None] = None,
        charge: int | list[int | None] = None,
        molecular_weight: float | list[float | None] = None,
        density: float | list[float | None] = None,
    ) -> None:
        """
        Initialize Component.

        Parameters
        ----------
        name : str | None
            Name of the component.
        species : str | list[str | None]
            Name(s) of the subspecies to initialize. If None, the component's name is used.
        charge : int | list [int | None]
            Charges of the subspecies. Defaults to None for each species.
        molecular_weight : float | list[float | None]
            Molecular weights of the subspecies. Defaults to None for each species.
        density : float | list[float | None]
            Density of component (including species). Defaults to None for each species.
        """
        self.name: str | None = name
        self._species: list[Species] = []

        if species is None:
            self.add_species(name, charge, molecular_weight, density)
        elif isinstance(species, str):
            self.add_species(species, charge, molecular_weight, density)
        elif isinstance(species, list):
            if charge is None:
                charge = len(species) * [None]
            if molecular_weight is None:
                molecular_weight = len(species) * [None]
            if density is None:
                density = len(species) * [None]
            for i, spec in enumerate(species):
                self.add_species(spec, charge[i], molecular_weight[i], density[i])
        else:
            raise CADETProcessError("Could not determine number of species")

    @property
    def species(self) -> list[Species]:
        """list[Species]: The subspecies of the component."""
        return self._species

    @wraps(Species.__init__)
    def add_species(
        self,
        species: str | Species,
        *args: Any,
        **kwargs: Any,
    ) -> Species:
        """
        Add a subspecies to the component.

        Parameters
        ----------
        species: string | Species
            Species to add
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.

        Returns
        -------
        Species
            The subspecies that was added.
        """
        if not isinstance(species, Species):
            species = Species(species, *args, **kwargs)
        self._species.append(species)
        return species

    @property
    def n_species(self) -> int:
        """int: The number of subspecies in the component."""
        return len(self.species)

    @property
    def label(self) -> list[str]:
        """list[str]: The names of the subspecies."""
        return [spec.name for spec in self.species]

    @property
    def charge(self) -> list[int | None]:
        """list[int | None]: The charges of the subspecies."""
        return [spec.charge for spec in self.species]

    @property
    def molecular_weight(self) -> list[float | None]:
        """list[float | None]: The molecular weights of the subspecies."""
        return [spec.molecular_weight for spec in self.species]

    @property
    def density(self) -> list[float | None]:
        """list[float | None]: The density of the subspecies."""
        return [spec.density for spec in self.species]

    def __str__(self) -> str:
        """str: String representation of the component."""
        return self.name

    def __iter__(self) -> Iterator[Species]:
        """Iterate over the subspecies of the component."""
        yield from self.species


class ComponentSystem(Structure):
    """
    Information about components in system.

    A component can contain subspecies (e.g. differently charged variants).

    Attributes
    ----------
    name : String
        Name of the component system.
    components : list[Component]
        List of individual components.
    n_species : int
        Number of Subspecies.
    n_comp : int
        Number of all component species.
    n_components : int
        Number of components.
    indices : dict[str, list[int]]
        Component indices.
    names : list[str]
        Names of all components.
    species : list[str]
        Names of all component species.
    charges : list[int | None]
        Charges of all components species.
    molecular_weights : list[float | None]
        Molecular weights of all component species.
    densities : list[float | None]
        Densities of all component species.

    See Also
    --------
    Species
    Component
    """

    name: String = String()

    def __init__(
        self,
        components: int | list[str | Component | None] = None,
        name: str | None = None,
        charges: list[int | None] = None,
        molecular_weights: list[float | None] = None,
        densities: list[float | None] = None,
    ) -> None:
        """
        Initialize the ComponentSystem object.

        Parameters
        ----------
        components : int | list[str | Component | None]
            The number of components or the list of components to be added.
            If None, no components are added.
        name : str | None
            The name of the ComponentSystem.
        charges : list[int | None]
            The charges of each component.
        molecular_weights : list[float | None]
            The molecular weights of each component.
        densities : list[float | None]
            The densities of each component.

        Raises
        ------
        CADETProcessError
            If the `components` argument is neither an int nor a list.
        """
        self.name: str | None = name
        self._components: list[Component] = []

        if components is None:
            return

        if isinstance(components, int):
            n_comp = components
            components = [str(i) for i in range(n_comp)]
        elif isinstance(components, list):
            n_comp = len(components)
        else:
            raise CADETProcessError("Could not determine number of components")

        if charges is None:
            charges = n_comp * [None]
        if molecular_weights is None:
            molecular_weights = n_comp * [None]
        if densities is None:
            densities = n_comp * [None]

        for i, comp in enumerate(components):
            self.add_component(
                comp,
                charge=charges[i],
                molecular_weight=molecular_weights[i],
                density=densities[i],
            )

    @property
    def components(self) -> list[Component]:
        """list[Component]: List of components in the system."""
        return self._components

    @property
    def components_dict(self) -> dict[str, Component]:
        """dict[str, Component]: Components indexed by name."""
        return {name: comp for name, comp in zip(self.names, self.components)}

    @property
    def n_components(self) -> int:
        """int: Number of components."""
        return len(self.components)

    @property
    def n_comp(self) -> int:
        """int: Number of species."""
        return self.n_species

    @property
    def n_species(self) -> int:
        """int: Number of species."""
        return sum([comp.n_species for comp in self.components])

    @wraps(Component.__init__)
    def add_component(
        self,
        component: str | Component,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Add a component to the system.

        Parameters
        ----------
        component : str | Component
            The component instance or name of the component to be added.
        *args : Any
            The positional arguments to be passed to the component class's constructor.
        **kwargs : Any
            The keyword arguments to be passed to the component class's constructor.
        """
        if not isinstance(component, Component):
            component = Component(component, *args, **kwargs)

        if component.name in self.names:
            raise CADETProcessError(
                f"Component '{component.name}' already exists in ComponentSystem."
            )

        self._components.append(component)

    def remove_component(self, component: str | Component) -> None:
        """
        Remove a component from the system.

        Parameters
        ----------
        component : str | Component
            The name of the component or the component instance to be removed.

        Raises
        ------
        CADETProcessError
            If the component is unknown or not present in the system.
        """
        if isinstance(component, str):
            try:
                component = self.components_dict[component]
            except KeyError:
                raise CADETProcessError("Unknown Component.")

        if component not in self.components:
            raise CADETProcessError("Unknown Component.")

        self._components.remove(component)

    @property
    def indices(self) -> dict[str, list[int]]:
        """dict[str, list[int]]: List of species indices for each component name."""
        indices = defaultdict(list)

        index = 0
        for comp in self.components:
            for spec in comp.species:
                indices[comp.name].append(index)
                index += 1

        return Dict(indices)

    @property
    def species_indices(self) -> dict[str, int]:
        """dict[str, int]: Indices for each species."""
        indices = Dict()

        index = 0
        for comp in self.components:
            for spec in comp.species:
                indices[spec.name] = index
                index += 1

        return indices

    @property
    def names(self) -> list[str]:
        """list[str]: List of component names."""
        names = [
            comp.name if comp.name is not None else str(i)
            for i, comp in enumerate(self.components)
        ]

        return names

    @property
    def species(self) -> list[str]:
        """list[str]: List of species names."""
        species = []
        index = 0
        for comp in self.components:
            for label in comp.label:
                if label is None:
                    species.append(str(index))
                else:
                    species.append(label)

                index += 1

        return species

    @property
    def charges(self) -> list[int | None]:
        """list[int | None]: List of species charges."""
        charges = []
        for comp in self.components:
            charges += comp.charge

        return charges

    @property
    def molecular_weights(self) -> list[float | None]:
        """list[float | None]: List of species molecular weights."""
        molecular_weights = []
        for comp in self.components:
            molecular_weights += comp.molecular_weight

        return molecular_weights

    @property
    def densities(self) -> list[float | None]:
        """list[float | None]: List of species densities."""
        densities = []
        for comp in self.components:
            densities += comp.density

        return densities

    def __repr__(self) -> str:
        """str: Return the string representation of the object."""
        return f"{self.__class__.__name__}({self.names})"

    def __len__(self) -> int:
        """int: Return the number of components in the system."""
        return self.n_comp

    def __iter__(self) -> Iterator[Component]:
        """Iterate over components in the system."""
        yield from self.components

    def __getitem__(self, item: int) -> Component:
        """Component: Retrieve a component by its index."""
        return self._components[item]



import copy

import numpy as np
import pytest
from CADETProcess.processModel import ComponentSystem, FlowSheet, Inlet, Process
from examples.batch_elution.process import process as batch_elution_process

TEST_PROFILE_CYCLE_TIME = 155 * 60


def new_inlet_only_process() -> Process:
    """Create a new inlet-only process for testing."""
    component_system = ComponentSystem(1)
    flow_sheet = FlowSheet(component_system)
    inlet = Inlet(component_system, "inlet")
    flow_sheet.add_unit(inlet)
    process = Process(flow_sheet, "inlet_only")
    process.cycle_time = TEST_PROFILE_CYCLE_TIME
    return process


def new_batch_elution_process() -> Process:
    """Create a deep copy of the batch elution process for testing."""
    return copy.deepcopy(batch_elution_process)


@pytest.fixture(params=["inlet_only"])
def process_fixture(request) -> Process:
    """
    Pytest fixture that initializes a fresh process for every test.

    Supports `inlet_only` (default) and `batch_elution` (for future tests).
    """
    if request.param == "inlet_only":
        return new_inlet_only_process()
    elif request.param == "batch_elution":
        return new_batch_elution_process()
    else:
        raise ValueError(f"Unknown process type: {request.param}")


def derivative_of_negative_gaussian(
    t: np.ndarray,
    center: float,
    sigma: float,
    amplitude: float,
    constant_value: float,
) -> np.ndarray:
    """
    Compute the derivative of a negative Gaussian function.

    Used for generating flow rate profiles.
    """
    neg_gauss = -amplitude * np.exp(-((t - center) ** 2) / (2 * sigma**2))
    neg_gauss_derivative = -((t - center) / (sigma**2)) * neg_gauss
    return neg_gauss_derivative + constant_value


def generate_input_profile(
    profile_type: str = "gaussian",
    duration: float = TEST_PROFILE_CYCLE_TIME,
    n_points: int = 1,
    **kwargs,
) -> tuple[np.ndarray, np.ndarray]:
    """Generate a time and flow rate profile for different test cases."""
    if profile_type == "gaussian":
        time_points = np.linspace(0, duration, n_points)
        pulse_center = kwargs.get("pulse_center", duration * 0.6)
        sigma = kwargs.get("sigma", duration * 0.1)
        amplitude = kwargs.get("amplitude", 1e-10)
        constant_value = kwargs.get("constant_value", 1e-5)
        input_profile = derivative_of_negative_gaussian(
            time_points, pulse_center, sigma, amplitude, constant_value
        )

    elif profile_type == "constant":
        time_points = np.linspace(0, duration, n_points)
        input_profile = np.full_like(time_points, kwargs.get("value", 1e-5))

    elif profile_type == "linear":
        time_points = np.linspace(0, duration, n_points)
        input_profile = np.linspace(
            kwargs.get("start", -1e-5), kwargs.get("end", 1e-5), n_points
        )

    elif profile_type == "two_point":
        time_points = np.array(
            [0, 1]
        )  # At least two points to prevent interpolation errors
        input_profile = np.array([kwargs.get("value", 1e-5), kwargs.get("value", 1e-5)])

    else:
        raise ValueError(f"Unknown profile type: {profile_type}")

    return time_points, input_profile


def test_add_flow_rate_profile_negative_gaussian(process_fixture: Process) -> None:
    """Test adding a negative Gaussian flow rate profile."""
    process = process_fixture
    time_points, input_profile = generate_input_profile(
        "gaussian", n_points=TEST_PROFILE_CYCLE_TIME
    )

    process.add_flow_rate_profile("inlet", time_points, input_profile)

    assert len(process.events) > 0, "No events were added to the process"
    assert np.allclose(
        input_profile,
        process.parameter_timelines["flow_sheet.inlet.flow_rate"].value(time_points),
        rtol=1e-2,
        atol=1e-8,
    ), "Event does not match the expected flow rate profile."


def test_add_flow_rate_profile_two_point(process_fixture: Process) -> None:
    """Test adding a flow rate profile with only two points."""
    process = process_fixture
    time_points, input_profile = generate_input_profile("two_point")

    process.add_flow_rate_profile(
        "inlet", time_points, input_profile, interpolation_method="pchip"
    )

    assert len(process.events) > 0, "No events were added for a two-point profile"
    assert np.allclose(
        input_profile,
        process.parameter_timelines["flow_sheet.inlet.flow_rate"].value(time_points),
        rtol=1e-2,
        atol=1e-8,
    ), "Event does not match the expected flow rate profile."


def test_add_flow_rate_profile_constant_function(process_fixture: Process) -> None:
    """Test adding a constant flow rate profile."""
    process = process_fixture
    time_points, input_profile = generate_input_profile("constant", n_points=50)

    process.add_flow_rate_profile(
        "inlet", time_points, input_profile, interpolation_method="pchip"
    )

    assert len(process.events) > 0, "No events were added for a constant function"
    assert np.allclose(
        input_profile,
        process.parameter_timelines["flow_sheet.inlet.flow_rate"].value(time_points),
        rtol=1e-2,
        atol=1e-8,
    ), "Event does not match the expected flow rate profile."


def test_add_flow_rate_profile_unordered_time(process_fixture: Process) -> None:
    """Test exception when time points are unordered."""
    process = process_fixture
    time_points = np.array([10, 0, 20, 5])
    input_profile = np.array([1e-5, 1e-5, 1e-5, 1e-5])

    with pytest.raises(ValueError, match="`x` must be strictly increasing sequence."):
        process.add_flow_rate_profile(
            "inlet", time_points, input_profile, interpolation_method="pchip"
        )


@pytest.mark.parametrize("method", ["cubic", "pchip", None])
def test_add_flow_rate_profile_different_interpolation_methods(
    process_fixture: Process, method: str
) -> None:
    """Test different interpolation methods for adding a flow rate profile."""
    process = process_fixture
    time_points, input_profile = generate_input_profile(
        "gaussian", n_points=TEST_PROFILE_CYCLE_TIME
    )

    process.add_flow_rate_profile(
        "inlet", time_points, input_profile, interpolation_method=method
    )

    assert len(process.events) > 0, (
        f"No events were added for interpolation method {method}"
    )
    assert np.allclose(
        input_profile,
        process.parameter_timelines["flow_sheet.inlet.flow_rate"].value(time_points),
        rtol=1e-2,
        atol=1e-8,
    ), f"Event does not match the expected flow rate profile for method {method}."



"""
=================================================
Fractionation (:mod:`CADETProcess.fractionation`)
=================================================

.. currentmodule:: CADETProcess.fractionation

A module to calculate process performance indicators.

.. autosummary::
   :toctree: generated/

   Fraction
   FractionPool
   Fractionator
   FractionationOptimizer

"""

from .fractions import *
from .fractionator import *
from .fractionationOptimizer import *



import warnings
from copy import deepcopy
from functools import wraps
from typing import Any, NoReturn, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from addict import Dict
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from CADETProcess import CADETProcessError, SimulationResults, plotting
from CADETProcess.dataStructure import (
    Integer,
    Structure,
    UnsignedFloat,
    UnsignedInteger,
)
from CADETProcess.processModel import (
    BindingBaseClass,
    ComponentSystem,
    Cstr,
    FlowSheet,
    Inlet,
    Langmuir,
    Linear,
    Outlet,
    Process,
    TubularReactor,
    TubularReactorBase,
    UnitBaseClass,
)
from CADETProcess.solution import SolutionBase

__all__ = [
    "SerialZone",
    "ParallelZone",
    "CarouselBuilder",
    "SMBBuilder",
    "LinearSMBBuilder",
    "LangmuirSMBBuilder",
]


class ZoneBaseClass(UnitBaseClass):
    """
    Base class for a multi-column zone with configurable columns and flow directions.

    Attributes
    ----------
    n_columns : UnsignedInteger
        The number of columns in the zone.
    flow_direction : Integer
        The flow direction in the zone, where 1 indicates normal and -1 indicates
        reverse flow.
    """

    n_columns = UnsignedInteger()
    flow_direction = Integer(default=1)

    def __init__(
        self,
        component_system: ComponentSystem,
        name: str,
        n_columns: int = 1,
        flow_direction: int = 1,
        initial_state: Optional[list] = None,
        valve_parameters: Optional[dict] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a ZoneBaseClass instance.

        Parameters
        ----------
        component_system : Any
            The component system associated with this zone.
        name : str
            The name of the zone.
        n_columns : int, optional
            Number of columns in the zone.
        flow_direction : int, optional
            Flow direction in the zone.
        initial_state : list, optional
            Initial state of the zone.
        valve_parameters : dict
            Additional parameters to setup mixer/splitter valves.
        *args : Optional
            Additional Parameters passed down to UnitBaseClass
        **kwargs : Optional
            Additional Parameters passed down to UnitBaseClass
        """
        self.n_columns = n_columns
        self.flow_direction = flow_direction
        self.initial_state = initial_state

        valve_parameters = valve_parameters or {}

        self._inlet_unit = self._setup_valve(
            component_system, f"{name}_inlet", **valve_parameters
        )
        self._outlet_unit = self._setup_valve(
            component_system, f"{name}_outlet", **valve_parameters
        )

        super().__init__(component_system, name, *args, **kwargs)

    @property
    def initial_state(self) -> list[dict[str, list]]:
        """list: The initial state of the columns in the zone."""
        return self._initial_state

    @initial_state.setter
    def initial_state(
        self,
        initial_state: list[dict[str, list]] | dict[str, list],
    ) -> NoReturn:
        if initial_state is None:
            self._initial_state = initial_state
            return

        if not isinstance(initial_state, list):
            initial_state = self.n_columns * [initial_state]

        if len(initial_state) != self.n_columns:
            raise CADETProcessError(f"Expected size {self.n_columns}")

        self._initial_state = initial_state

    def _setup_valve(
        self,
        component_system: ComponentSystem,
        name: str,
        unit_type: Union[Cstr, TubularReactor] = Cstr,
        **valve_parameters: dict,
    ) -> None:
        if unit_type not in (Cstr, TubularReactor):
            raise ValueError(
                "Unknown unit type. Must be one of `Cstr`, `TubularReactor`."
            )

        valve = unit_type(component_system, name)

        valve_dead_volume = valve_parameters.get("valve_dead_volume", 1e-9)

        if isinstance(valve, Cstr):
            valve.init_liquid_volume = valve_dead_volume
        elif isinstance(valve, TubularReactor):
            valve.discretization.ncol = 1

            length_diameter_ratio = valve_parameters.get("length_diameter_ratio", 1)
            valve.diameter = (
                4 / np.pi * valve_dead_volume / length_diameter_ratio
            ) ** (1 / 3)
            valve.length = length_diameter_ratio * valve.diameter

            valve.axial_dispersion = valve_parameters.get("axial_dispersion", 0)

        return valve

    @property
    def inlet_unit(self) -> Cstr:
        """Cstr: The inlet CSTR unit of the zone."""
        return self._inlet_unit

    @property
    def outlet_unit(self) -> Cstr:
        """Cstr: The outlet CSTR unit of the zone."""
        return self._outlet_unit


class SerialZone(ZoneBaseClass):
    """Zone with columns connected in series."""

    pass


class ParallelZone(ZoneBaseClass):
    """Zone with columns connected in parallel."""

    pass


class CarouselBuilder(Structure):
    """
    Configurator for multi-column processes.

    Attributes
    ----------
    component_system : Any
        The system of components for which the carousel is configured.
    name : str
        Name of the carousel system.
    switch_time : float
        Column switch time.
    valve_parameters : dict
        Additional parameters to setup mixer/splitter valves.
    """

    switch_time = UnsignedFloat()

    def __init__(
        self,
        component_system: ComponentSystem,
        name: str,
        valve_parameters: dict = None,
    ) -> NoReturn:
        """
        Initialize a CarouselBuilder instance.

        Parameters
        ----------
        component_system : Any
            The system of components that will be used in the carousel.
        name : str
            The carousel name.
        valve_parameters : dict
            Additional parameters to setup mixer/splitter valves.
        """
        self.component_system = component_system
        self.name = name
        self._flow_sheet = FlowSheet(component_system, name)
        self._column = None
        self.valve_parameters = valve_parameters

    @property
    def flow_sheet(self) -> FlowSheet:
        """FlowSheet: The flow sheet instance associated with the carousel."""
        return self._flow_sheet

    @property
    def column(self) -> TubularReactorBase:
        """TubularReactorBase: The column template for all zones."""
        return self._column

    @column.setter
    def column(self, column: TubularReactorBase) -> NoReturn:
        if not isinstance(column, TubularReactorBase):
            raise TypeError("Column must be an instance of TubularReactorBase.")
        if self.component_system is not column.component_system:
            raise CADETProcessError("Number of components does not match.")
        self._column = column

    @wraps(FlowSheet.add_unit)
    def add_unit(self, *args: Any, **kwargs: Any) -> None:
        """Wrap FlowSheet.add_unit to add a unit to the flow sheet."""
        self.flow_sheet.add_unit(*args, **kwargs)

    @wraps(FlowSheet.add_connection)
    def add_connection(self, *args: Any, **kwargs: Any) -> None:
        """Wrap FlowSheet.add_connection to add a connection between units."""
        self.flow_sheet.add_connection(*args, **kwargs)

    @wraps(FlowSheet.set_output_state)
    def set_output_state(self, *args: Any, **kwargs: Any) -> None:
        """Wrap FlowSheet.set_output_state to set the output state of a unit."""
        self.flow_sheet.set_output_state(*args, **kwargs)

    @property
    def zones(self) -> list[ZoneBaseClass]:
        """
        Get all zones in the carousel system.

        Returns
        -------
        list
            A list of all zones in the carousel system.
        """
        return [
            unit for unit in self.flow_sheet.units if isinstance(unit, ZoneBaseClass)
        ]

    @property
    def zone_names(self) -> list[str]:
        """list: Zone names."""
        return [zone.name for zone in self.zones]

    @property
    def zones_dict(self) -> dict[str, ZoneBaseClass]:
        """dict: Zone names and objects."""
        return {zone.name: zone for zone in self.zones}

    @property
    def n_zones(self) -> int:
        """int: Number of zones in the Carousel System."""
        return len(self.zones)

    @property
    def n_columns(self) -> int:
        """int: Number of columns in the Carousel System."""
        return sum([zone.n_columns for zone in self.zones])

    def build_flow_sheet(self) -> FlowSheet:
        """
        Assemble the flow sheet.

        Returns
        -------
        FlowSheet
            The assembled flow sheet.
        """
        if self.column is None:
            raise CADETProcessError("No column associated with Carousel.")

        flow_sheet = FlowSheet(self.component_system, self.name)

        self._add_units(flow_sheet)
        self._add_inter_zone_connections(flow_sheet)
        self._add_intra_zone_connections(flow_sheet)
        self._set_output_states(flow_sheet)

        return flow_sheet

    def _add_units(self, flow_sheet: FlowSheet) -> None:
        """Add units to flow_sheet."""
        col_index = 0
        for unit in self.flow_sheet.units:
            if not isinstance(unit, ZoneBaseClass):
                is_feed_inlet = unit in self.flow_sheet.feed_inlets
                is_eluent_inlet = unit in self.flow_sheet.eluent_inlets
                is_output_outlet = unit in self.flow_sheet.product_outlets
                flow_sheet.add_unit(
                    unit,
                    feed_inlet=is_feed_inlet,
                    eluent_inlet=is_eluent_inlet,
                    product_outlet=is_output_outlet,
                )
            else:
                flow_sheet.add_unit(unit.inlet_unit)
                flow_sheet.add_unit(unit.outlet_unit)
                for i_col in range(unit.n_columns):
                    col = deepcopy(self.column)
                    col.component_system = self.component_system
                    col.name = f"column_{col_index}"
                    if unit.initial_state is not None:
                        col.initial_state = unit.initial_state[i_col]
                    flow_sheet.add_unit(col)
                    col_index += 1

    def _add_inter_zone_connections(self, flow_sheet: FlowSheet) -> NoReturn:
        """Add connections between zones."""
        for unit, connections in self.flow_sheet.connections.items():
            if isinstance(unit, ZoneBaseClass):
                origin = unit.outlet_unit
            else:
                origin = unit
            if connections.destinations:
                for destination in connections.destinations[None]:
                    if isinstance(destination, ZoneBaseClass):
                        destination = destination.inlet_unit

                    flow_sheet.add_connection(origin, destination)

    def _add_intra_zone_connections(self, flow_sheet: FlowSheet) -> NoReturn:
        """Add connections within zones."""
        for zone in self.zones:
            for col_index in range(self.n_columns):
                col = flow_sheet[f"column_{col_index}"]
                flow_sheet.add_connection(zone.inlet_unit, col)
                col = flow_sheet[f"column_{col_index}"]
                flow_sheet.add_connection(col, zone.outlet_unit)

        for col_index in range(self.n_columns):
            col_orig = flow_sheet[f"column_{col_index}"]
            if col_index < self.n_columns - 1:
                col_dest = flow_sheet[f"column_{col_index + 1}"]
            else:
                col_dest = flow_sheet[f"column_{0}"]
            flow_sheet.add_connection(col_orig, col_dest)

    def _set_output_states(self, flow_sheet: FlowSheet) -> NoReturn:
        for unit in self.flow_sheet.output_states:
            output_state = self.flow_sheet.output_states[unit]

            if output_state == {}:
                continue

            if isinstance(unit, ZoneBaseClass):
                flow_sheet.set_output_state(unit.outlet_unit, output_state)
            else:
                flow_sheet.set_output_state(unit, output_state)

    def build_process(self) -> Process:
        """
        Assemble the process object.

        Returns
        -------
        Process
            The assembled process object.
        """
        flow_sheet = self.build_flow_sheet()
        process = Process(flow_sheet, self.name)

        self._add_events(process)

        return process

    @property
    def cycle_time(self) -> float:
        """float: cycle time of the process."""
        return self.n_columns * self.switch_time

    def _add_events(self, process: Process) -> None:
        """Add events to process."""
        process.cycle_time = self.n_columns * self.switch_time
        process.add_duration("switch_time", self.switch_time)

        for carousel_state in range(self.n_columns):
            position_counter = 0
            for i_zone, zone in enumerate(self.zones):
                col_indices = np.arange(zone.n_columns)
                col_indices += position_counter
                col_indices = self.column_indices_at_state(col_indices, carousel_state)

                if isinstance(zone, SerialZone):
                    evt = process.add_event(
                        f"{zone.name}_{carousel_state}",
                        f"flow_sheet.output_states.{zone.inlet_unit}",
                        col_indices[0],
                    )
                    process.add_event_dependency(
                        evt.name, "switch_time", [carousel_state]
                    )
                    for i, col in enumerate(col_indices):
                        if i < (zone.n_columns - 1):
                            evt = process.add_event(
                                f"column_{col}_{carousel_state}",
                                f"flow_sheet.output_states.column_{col}",
                                self.n_zones,
                            )
                        else:
                            evt = process.add_event(
                                f"column_{col}_{carousel_state}",
                                f"flow_sheet.output_states.column_{col}",
                                i_zone,
                            )
                        process.add_event_dependency(
                            evt.name, "switch_time", [carousel_state]
                        )
                elif isinstance(zone, ParallelZone):
                    output_state = self.n_columns * [0]
                    for col in col_indices:
                        output_state[col] = 1 / zone.n_columns

                    evt = process.add_event(
                        f"{zone.name}_{carousel_state}",
                        f"flow_sheet.output_states.{zone.inlet_unit}",
                        output_state,
                    )
                    process.add_event_dependency(
                        evt.name, "switch_time", [carousel_state]
                    )

                    for col in col_indices:
                        evt = process.add_event(
                            f"column_{col}_{carousel_state}",
                            f"flow_sheet.output_states.column_{col}",
                            i_zone,
                        )
                        process.add_event_dependency(
                            evt.name, "switch_time", [carousel_state]
                        )

                for i, col in enumerate(col_indices):
                    evt = process.add_event(
                        f"column_{col}_{carousel_state}_velocity",
                        f"flow_sheet.column_{col}.flow_direction",
                        zone.flow_direction,
                    )
                    process.add_event_dependency(
                        evt.name, "switch_time", [carousel_state]
                    )

                position_counter += zone.n_columns

    def carousel_state(self, t: float) -> int:
        """
        Return carousel state at given time.

        Parameters
        ----------
        t: float
            Time

        Returns
        -------
        int:
             Carousel state at given time.
        """
        return int(np.floor((t % self.cycle_time) / self.switch_time))

    def column_indices_at_state(
        self,
        carousel_positions: np.typing.NDArray[int],
        carousel_state: int,
    ) -> np.ndarray[int]:
        """
        Determine index of column unit at given carousel position and state.

        Parameters
        ----------
        carousel_positions: np.typing.NDArray[np.int_]
            Carousel position indices (e.g. wash position, elute position).
        carousel_state : int
            Curent state of the carousel system.

        Returns
        -------
        np.ndarray[int]
            Indices of column units at given carousel positions and state.
        """
        carousel_positions = np.array(carousel_positions, dtype=int)

        return (carousel_positions + carousel_state) % self.n_columns

    def column_indices_at_time(
        self,
        t: float,
        carousel_positions: np.typing.NDArray[int],
    ) -> int:
        """
        Determine index of column unit at given carousel position and time.

        Parameters
        ----------
        t: float
            Time
        carousel_position : int
            Carousel position.

        Returns
        -------
        np.ndarray[int]
            Indices of column units at given carousel positions and time.
        """
        carousel_positions = np.array(carousel_positions, dtype=int)

        carousel_state = self.carousel_state(t)
        column_indices = self.column_indices_at_state(
            carousel_positions, carousel_state
        )

        return column_indices


class SMBBuilder(CarouselBuilder):
    """
    Configurator for 4 Zone SMB systems.

    Attributes
    ----------
    binding_model_type : Type[BindingBaseClass]
        Specifies the type of binding model used in the SMB process.
    """

    binding_model_type = BindingBaseClass

    def __init__(
        self,
        feed: Inlet,
        eluent: Inlet,
        column: TubularReactorBase,
        name: str = "SMB",
        valve_parameters: dict = None,
    ) -> NoReturn:
        """
        Initialize an SMBBuilder instance.

        Parameters
        ----------
        feed : Inlet
            The feed inlet to the SMB system.
        eluent : Inlet
            The eluent inlet to the SMB system.
        column : TubularReactorBase
            The column equipment in the SMB system.
        name : str, optional
            The name of the SMB setup, by default 'SMB'.

        Raises
        ------
        CADETProcessError
            If component systems of feed, eluent, and column do not match.
        TypeError
            If feed, eluent, or column are not of the expected type.
        """
        component_system = feed.component_system
        if not (component_system is eluent.component_system is column.component_system):
            raise CADETProcessError("ComponentSystems do not match.")
        if not isinstance(feed, Inlet):
            raise TypeError(f"Expected inlet object. Got {type(feed)}.")
        if not isinstance(eluent, Inlet):
            raise TypeError(f"Expected inlet object. Got {type(eluent)}.")

        if not isinstance(column, TubularReactorBase):
            raise TypeError(f"Expected Column object. Got {type(column)}.")

        self._validate_binding_model(column.binding_model)

        super().__init__(component_system, name, valve_parameters)

        raffinate = Outlet(component_system, name="raffinate")
        extract = Outlet(component_system, name="extract")

        zone_I = SerialZone(
            component_system, "zone_I", 1, valve_parameters=self.valve_parameters
        )
        zone_II = SerialZone(
            component_system, "zone_II", 1, valve_parameters=self.valve_parameters
        )
        zone_III = SerialZone(
            component_system, "zone_III", 1, valve_parameters=self.valve_parameters
        )
        zone_IV = SerialZone(
            component_system, "zone_IV", 1, valve_parameters=self.valve_parameters
        )

        # Carousel Builder
        self.column = column
        self.add_unit(feed, feed_inlet=True)
        self.add_unit(eluent, eluent_inlet=True)

        self.add_unit(raffinate, product_outlet=True)
        self.add_unit(extract, product_outlet=True)

        self.add_unit(zone_I)
        self.add_unit(zone_II)
        self.add_unit(zone_III)
        self.add_unit(zone_IV)

        self.add_connection(eluent, zone_I)

        self.add_connection(zone_I, extract)
        self.add_connection(zone_I, zone_II)

        self.add_connection(zone_II, zone_III)

        self.add_connection(feed, zone_III)

        self.add_connection(zone_III, raffinate)
        self.add_connection(zone_III, zone_IV)

        self.add_connection(zone_IV, zone_I)

    def _validate_binding_model(self, binding_model: BindingBaseClass) -> None:
        """
        Validate that the provided binding model matches the required type.

        Parameters
        ----------
        binding_model : BindingBaseClass
            The binding model to be validated.

        Raises
        ------
        TypeError
            If the binding model does not match the expected type.
        """
        if not isinstance(binding_model, self.binding_model_type):
            raise TypeError(
                f"Invalid binding model. Expected {self.binding_model_type}."
            )

    def _get_zone_flow_rates(self, m: list, switch_time: float) -> list[float]:
        m1, m2, m3, m4 = m

        Vc = self.column.volume
        et = self.column.total_porosity

        Q_I = Vc * (m1 * (1 - et) + et) / switch_time     # Flow rate Zone I
        Q_II = Vc * (m2 * (1 - et) + et) / switch_time    # Flow rate Zone II
        Q_III = Vc * (m3 * (1 - et) + et) / switch_time   # Flow rate Zone III
        Q_IV = Vc * (m4 * (1 - et) + et) / switch_time    # Flow rate Zone IV

        return [Q_I, Q_II, Q_III, Q_IV]

    def _get_unit_flow_rates(self, Q_zones: list[float]) -> list[float]:
        Q_I, Q_II, Q_III, Q_IV = Q_zones

        Q_feed = Q_III - Q_II
        Q_eluent = Q_I - Q_IV
        Q_raffinate = Q_III - Q_IV
        Q_extract = Q_I - Q_II

        return [Q_feed, Q_eluent, Q_raffinate, Q_extract]

    def _get_split_ratios(
        self, Q_zones: list[float], Q_units: list[float]
    ) -> tuple[float, float]:
        Q_I, Q_II, Q_III, Q_IV = Q_zones
        Q_feed, Q_eluent, Q_raffinate, Q_extract = Q_units

        w_r = Q_raffinate / Q_III
        w_e = Q_extract / Q_I

        return w_r, w_e

    def get_design_parameters(
        self,
        binding_model: BindingBaseClass,
        c_feed: np.ndarray,
    ) -> Any:
        """
        Retrieve design parameters based on the binding model.

        Parameters
        ----------
        binding_model : BindingBaseClass
            The binding model used to calculate design parameters.
        c_feed : np.ndarray
            The feed concentration.

        Returns
        -------
        Any
            The design parameters computed from the binding model.

        Raises
        ------
        NotImplementedError
            This method should be implemented by subclasses.
        """
        raise NotImplementedError("Subclass must implement this method.")

    def calculate_m_opt(self, *design_parameters: Any) -> list:
        """
        Calculate the optimal zone flow rates based on provided design parameters.

        Parameters
        ----------
        design_parameters : Any
            A variable number of parameters defining the design.

        Returns
        -------
        list
            The optimal zone flow rates based on the design parameters.

        Raises
        ------
        NotImplementedError
            This method should be implemented by subclasses.
        """
        raise NotImplementedError("Subclass must implement this method.")

    def apply_safety_factor(
        self,
        m_opt: list,
        *design_parameters: Any,
        gamma: float | list[float],
    ) -> Any:
        """
        Apply a safety factor to the optimal zone flow rates.

        Parameters
        ----------
        m_opt : Any
            The optimal zone flow rates calculated without safety considerations.
        design_parameters : Any
            A variable number of parameters defining the design.
        gamma : float | list[float]
            The safety factor(s) to apply to the zone flow rates. If float is provided,
            the same value is applied to all zone flow rates.

        Returns
        -------
        list
            The adjusted zone flow rates after applying the safety factor.

        Raises
        ------
        NotImplementedError
            This method should be implemented by subclasses.
        """
        raise NotImplementedError("Subclass must implement this method.")

    def triangle_design(
        self,
        binding_model: Optional[BindingBaseClass] = None,
        c_feed: Optional[np.ndarray] = None,
        switch_time: Optional[float] = None,
        gamma: float | list[float] = 1,
        set_values: bool = True,
    ) -> list[float]:
        """
        Design the SMB process according to the triangle theory.

        Parameters
        ----------
        binding_model : BindingBaseClass, optional
            The binding model to be used.
            If None, uses the binding model from the column attribute.
        c_feed : np.ndarray, optional
            Feed concentration matrix.
            If None, uses current feed concentrations.
        switch_time : float, optional
            Switching time for the SMB zones.
            If None, uses the existing switch_time attribute.
        gamma : float | list[float]
            The safety factor(s) to apply to the zone flow rates. If float is provided,
            the same value is applied to all zone flow rates.
            The default is 1.
        set_values : bool, default True
            If True, sets the calculated values to the process model.

        Returns
        -------
        list
            A list containing the feed flow rate, eluent flow rate, raffinate and
            extract split ratios.

        Raises
        ------
        Warning
            If binding_model is provided and set_values is True, values cannot be set.
        """
        if binding_model is None:
            binding_model = self.column.binding_model
        elif set_values is True:
            warnings.warn("Cannot set values if binding_model is given.")
            set_values = False

        self._validate_binding_model(binding_model)

        if c_feed is None:
            c_feed = self.flow_sheet.feed.c[:, 0]
        elif set_values is True:
            self.flow_sheet.feed.c = c_feed

        design_parameters = self.get_design_parameters(binding_model, c_feed)
        m_opt = self.calculate_m_opt(*design_parameters)
        m = self.apply_safety_factor(m_opt, *design_parameters, gamma=gamma)

        if switch_time is None:
            switch_time = self.switch_time
        elif set_values is True:
            self.switch_time = switch_time

        Q_zones = self._get_zone_flow_rates(m, switch_time)
        Q_units = self._get_unit_flow_rates(Q_zones)
        Q_feed, Q_eluent, Q_raffinate, Q_extract = Q_units
        w_r, w_e = self._get_split_ratios(Q_zones, Q_units)

        if set_values:
            self.flow_sheet.feed.flow_rate = Q_feed
            self.flow_sheet.eluent.flow_rate = Q_eluent
            self.set_output_state("zone_I", [w_e, 1 - w_e])
            self.set_output_state("zone_III", [w_r, 1 - w_r])

        return [Q_feed, Q_eluent, w_r, w_e]

    def plot_triangle(
        self,
        binding_model: Optional[BindingBaseClass] = None,
        c_feed: Optional[np.ndarray] = None,
        gamma: float | list[float] = 1,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
    ) -> tuple[plt.Figure, plt.Axes]:
        """
        Plot the triangle diagram for the SMB process with the operating point marked.

        Parameters
        ----------
        binding_model : BindingBaseClass, optional
            The binding model to use for the plot.
            If None, uses the column's binding model.
        c_feed : np.ndarray, optional
            The feed concentration array.
            If None, uses the current feed concentration from the flow sheet.
        gamma : float | list[float]
            The safety factor(s) to apply to the zone flow rates. If float is provided,
            the same value is applied to all zone flow rates.
        fig : plt.Figure, optional
            The matplotlib figure object. If None, a new figure will be created.
        ax : plt.Axes, optional
            The matplotlib axes object.
            If None, a new axes will be created in the new or specified figure.

        Returns
        -------
        tuple[plt.Figure, plt.Axes]
            The figure and axes objects containing the triangle diagram.

        Notes
        -----
        This method uses internal methods to fetch design parameters and calculate
        optimal zone flow rates, which should be defined in a subclass.
        """
        if binding_model is None:
            binding_model = self.column.binding_model
        self._validate_binding_model(binding_model)

        if c_feed is None:
            c_feed = self.flow_sheet.feed.c[:, 0]

        # Operating Point
        design_parameters = self.get_design_parameters(binding_model, c_feed)
        m_opt = self.calculate_m_opt(*design_parameters)
        m1, m2, m3, m4 = self.apply_safety_factor(
            m_opt, *design_parameters, gamma=gamma
        )

        # Setup figure
        if fig is None:
            fig, ax = plt.subplots(figsize=(10, 10))

        # Plot Triangle
        self._plot_triangle(ax, *design_parameters)

        ax.set_xlabel("$m_{II}$")
        ax.set_ylabel("$m_{III}$")

        # Operating point
        ax.scatter(m2, m3, c="k", marker="x", zorder=3)
        ax.annotate(
            "operating point",
            xy=(m2, m3),
            xytext=(1.1 * m2, 0.9 * m3),
            arrowprops=dict(facecolor="black", shrink=0.01),
        )

        return fig, ax

    def _plot_triangle(self, ax: Axes, *design_parameters: Any) -> None:
        """
        Plot a theoretical triangle diagram for SMB processes.

        This method needs to be implemented by subclasses to define the specific way the
        triangle is visualized based on provided design parameters.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The matplotlib axes on which the triangle will be plotted.
        design_parameters : Any
            Variable-length argument list of design parameters that define the triangle.

        Raises
        ------
        NotImplementedError
            This is a placeholder method and should be implemented in subclasses.
        """
        raise NotImplementedError()


class LinearSMBBuilder(SMBBuilder):
    """
    Configure a 4-zone SMB system with linear isotherms.

    Note, this is currently only supported for 2-Component systems.

    Attributes
    ----------
    binding_model_type : type
        Specifies that this builder uses the Linear binding model.
    """

    binding_model_type = Linear

    def _validate_binding_model(self, binding_model: BindingBaseClass) -> None:
        """
        Validate the binding model for compatibility with Langmuir SMB systems.

        Parameters
        ----------
        binding_model : BindingBaseClass
            The binding model to be validated.

        Raises
        ------
        CADETProcessError
            If the binding model does not contain exactly two components.

        Warns
        -----
        RuntimeWarning
            If the binding model is kinetic, which conflicts with the assumption of
            instant equilibrium.
        """
        super()._validate_binding_model(binding_model)

        if binding_model.n_comp != 2:
            raise CADETProcessError("This only works for 2-Component Systems.")

        if binding_model.is_kinetic:
            warnings.warn(
                "Isotherm uses kinetic binding, "
                "however, triangle theory assumes instant equilibrium."
            )

    def get_design_parameters(
        self,
        binding_model: BindingBaseClass,
        c_feed: np.ndarray,
    ) -> tuple[float, float]:
        """
        Calculate Henry's constants (H) based on adsorption and desorption rates.

        Parameters
        ----------
        binding_model : BindingBaseClass
            The binding model containing adsorption and desorption rate data.
        c_feed : np.ndarray
            The concentration feed.

        Returns
        -------
        Tuple[float, float]
            Henry's constants for the strongly and weakly adsorbing components.
        """
        k_ads = np.array(binding_model.adsorption_rate)
        k_des = np.array(binding_model.desorption_rate)
        H = k_ads / k_des

        if H[1] < H[0]:
            HA, HB = H
        else:
            HB, HA = H

        return HA, HB

    def calculate_m_opt(
        self,
        HA: float,
        HB: float,
    ) -> list[float]:
        """
        Calculate the optimal flow rates for SMB zones based on Henry's constants.

        Parameters
        ----------
        HA : float
            Henry's constant for strongly binding component.
        HB : float
            Henry's constant for weakly binding component.

        Returns
        -------
        list[float]
            List of optimal flow rates for zones 1 to 4, respectively.
        """
        m1 = HA
        m2 = HB
        m3 = HA
        m4 = HB

        return [m1, m2, m3, m4]

    def apply_safety_factor(
        self,
        m_opt: list[float],
        *design_parameters: Any,
        gamma: float | list[float] = 1,
    ) -> list[float]:
        """
        Adjust the optimal flow rates by applying safety factors to each zone.

        Parameters
        ----------
        m_opt : list[float]
            A list containing the optimal flow rates for zones 1 to 4.
        design_parameters : Union[float, List[float]]
            Additional design parameters that might influence safety factor application.
        gamma : float | list[float]
            Safety factors to apply. Can be a single float applied to all zones, or a
            list of floats for each zone. The default is 1.

        Returns
        -------
        list[float]
            Adjusted flow rates for zones 1 to 4 after applying the safety factors.

        Raises
        ------
        ValueError
            If the product of gamma for zones 2 and 3 is not smaller than the ratio of
            m3_opt to m2_opt, which is a critical operational constraint.
        """
        m1_opt, m2_opt, m3_opt, m4_opt = m_opt

        if np.isscalar(gamma):
            gamma = 4 * [gamma]

        gamma_1, gamma_2, gamma_3, gamma_4 = gamma

        if gamma_2 * gamma_3 >= m3_opt / m2_opt:
            raise ValueError("gamma_2 * gamma_3 must be smaller than HA / HB ")

        m1 = gamma_1 * m1_opt
        m2 = gamma_2 * m2_opt
        m3 = m3_opt / gamma_3
        m4 = m4_opt / gamma_4

        return [m1, m2, m3, m4]

    def _plot_triangle(
        self,
        ax: Axes,
        HA: float,
        HB: float,
    ) -> NoReturn:
        """
        Plot SMB triangle for linear isotherm.

        Notable points:
        - a: [HA, HA]
        - b: [HB, HB]
        - w_opt: [HB, HA]

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on.
        HA : float
            Henry coefficient of strongly binding component.
        HB : float
            Henry coefficient of strongly binding component.
        """
        # Bounds
        lb = HB - 0.3 * (HA - HB)
        ub = HA + 0.3 * (HA - HB)

        ax.set_xlim(lb, ub)
        ax.set_ylim(lb, ub)

        # Diagonal
        ax.plot((lb, ub), (lb, ub), "k")

        # Henry coefficients
        for h in [HB, HA]:
            ax.hlines(h, 0, h, "k", "dashed")
            ax.vlines(h, h, ub, "k", "dashed")

        # Triangle
        ax.hlines(HA, HB, HA, "k")
        ax.vlines(HB, HB, HA, "k")

        # Label regions
        ax.text(
            (HB + (HA - HB) / 2), (0.95 * ub), "Pure extract", ha="center", va="center"
        )
        ax.text(
            (1.05 * lb),
            (HB + (HA - HB) / 2),
            "Pure raffinate",
            ha="center",
            va="center",
            rotation="vertical",
        )


class LangmuirSMBBuilder(SMBBuilder):
    """
    Configure a 4-zone SMB system with Langmuir isotherms.

    Note, this is currently only supported for 2-Component systems.

    Attributes
    ----------
    binding_model_type : type
        Specifies that this builder uses the Linear binding model.
    """

    binding_model_type = Langmuir

    def _validate_binding_model(self, binding_model: BindingBaseClass) -> NoReturn:
        """
        Validate the binding model for compatibility with Langmuir SMB systems.

        Parameters
        ----------
        binding_model : BindingBaseClass
            The binding model to be validated.

        Raises
        ------
        CADETProcessError
            If the binding model does not contain exactly two components.

        Warns
        -----
        RuntimeWarning
            If the binding model is kinetic, which conflicts with the assumption of
            instant equilibrium.
        """
        super()._validate_binding_model(binding_model)

        if binding_model.n_comp != 2:
            raise CADETProcessError("This only works for 2-Component Systems.")

        if binding_model.is_kinetic:
            warnings.warn(
                "Isotherm uses kinetic binding, "
                "however, triangle theory assumes instant equilibrium."
            )

    def get_design_parameters(
        self,
        binding_model: BindingBaseClass,
        c_feed: np.ndarray,
    ) -> tuple[float, float]:
        """
        Calculate the optimal flow rates for SMB zones based on the provided parameters.

        Parameters
        ----------
        binding_model : BindingBaseClass
            The binding model parameters.
        c_feed : np.ndarray
            The concentration feed matrix.

        Returns
        -------
        tuple
            A tuple containing Henry's constants HA, HB, binding coefficients bA, bB,
            feed concentrations cFA, cFB, and quadratic nulls wG, wF.
        """
        k_ads = np.array(binding_model.adsorption_rate)
        k_des = np.array(binding_model.desorption_rate)
        k_eq = k_ads / k_des
        q_sat = np.array(binding_model.capacity)
        H = [k_eq * q_s for k_eq, q_s in zip(k_eq, q_sat)]

        if H[1] < H[0]:
            HA, HB = H
            bA, bB = k_eq
            cFA, cFB = c_feed
        else:
            HB, HA = H
            bB, bA = k_eq
            cFB, cFA = c_feed

        a = -(HA * (1 + bB * cFB) + HB * (1 + bA * cFA)) / (1 + bB * cFB + bA * cFA)
        b = HA * HB / (1 + bB * cFB + bA * cFA)
        wG = -a / 2 + np.sqrt((-a / 2) ** 2 - b)
        wF = -a / 2 - np.sqrt((-a / 2) ** 2 - b)

        return HA, HB, bA, bB, cFA, cFB, wG, wF

    def calculate_m_opt(
        self,
        HA: float,
        HB: float,
        bA: float,
        bB: float,
        cFA: float,
        cFB: float,
        wG: float,
        wF: float,
    ) -> list[float]:
        """
        Calculate optimal zone flow rates based on Langmuir isotherm parameters.

        Parameters
        ----------
        HA : float
            Henry's constant for strongly binding component.
        HB : float
            Henry's constant for weakly binding component.
        bA : float
            Binding coefficient for strongly binding component.
        bB : float
            Binding coefficient for weakly binding component.
        cFA : float
            Feed concentration for strongly binding component.
        cFB : float
            Feed concentration for weakly binding component.
        wG : float
            First quadratic null.
        wF : float
            Second quadratic null.

        Returns
        -------
        list[float]
            List of optimal flow rates for zones 1 to 4.
        """
        m1 = HA
        m2 = HB / HA * wG
        m3 = wG * (wF * (HA - HB) + HB * (HB - wF)) / (HB * (HA - wF))
        m4 = (
            1
            / 2
            * (
                HB
                + m3
                + bB * cFB * (m3 - m2)
                - np.sqrt((HB + m3 + bB * cFB * (m3 - m2)) ** 2 - 4 * HB * m3)
            )
        )

        return [m1, m2, m3, m4]

    def apply_safety_factor(
        self,
        m_opt: list[float],
        HA: float,
        HB: float,
        bA: float,
        bB: float,
        cFA: float,
        cFB: float,
        wG: float,
        wF: float,
        gamma: float | list[float] = 1,
    ) -> list[float]:
        """
        Apply a safety factor to the optimal zone flow rates.

        Parameters
        ----------
        m_opt : list[float]
            List of optimal flow rates for zones 1 to 4.
        HA : float
            Henry's constant for strongly binding component.
        HB : float
            Henry's constant for weakly binding component.
        bA : float
            Binding coefficient for strongly binding component.
        bB : float
            Binding coefficient for weakly binding component.
        cFA : float
            Feed concentration for strongly binding component.
        cFB : float
            Feed concentration for weakly binding component.
        wG : float
            First quadratic null.
        wF : float
            Second quadratic null.
        gamma : float | list[float]
            The safety factor(s) to apply to the zone flow rates. If float is provided,
            the same value is applied to all zone flow rates.

        Returns
        -------
        list[float]
            Adjusted flow rates for zones 1 to 4 after applying the safety factors.
        """
        m1_opt, m2_opt, m3_opt, m4_opt = m_opt

        if np.isscalar(gamma):
            W_opt = np.array([m2_opt, m3_opt])
            B = np.array([HB, HB])
            R = [
                wG**2 / HA,
                wG
                * (wF * (HA - wG) * (HA - HB) + HB * wG * (HA - wF))
                / (HA * HB * (HA - wF)),
            ]

            # Calculating vectors WB and WA
            WB = B - W_opt
            WR = R - W_opt

            # Normalizing vectors
            norm_WB = WB / np.linalg.norm(WB)
            norm_WR = WR / np.linalg.norm(WR)

            # Calculating the bisector WB / WR
            bisector = norm_WB + norm_WR
            norm_dir_vec_bisector = bisector / np.linalg.norm(bisector)

            # Calculating the new point W' using the safety factor gamma
            W = W_opt + (gamma - 1) * norm_dir_vec_bisector

            m1 = gamma * m1_opt
            m2 = W[0]
            m3 = W[1]
            m4 = m4_opt / gamma

        else:
            gamma_1, gamma_2, gamma_3, gamma_4 = gamma

            m1 = gamma_1 * m1_opt
            m2 = gamma_2 * m2_opt
            m3 = m3_opt / gamma_3
            m4 = m4_opt / gamma_4

        return [m1, m2, m3, m4]

    def _plot_triangle(
        self,
        ax: Axes,
        HA: float,
        HB: float,
        bA: float,
        bB: float,
        cFA: float,
        cFB: float,
        wG: float,
        wF: float,
    ) -> None:
        """
        Plot SMB triangle for Langmuir isotherm.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on.
        HA : float
            Henry's constant for strongly binding component.
        HB : float
            Henry's constant for weakly binding component.
        bA : float
            Binding coefficient for strongly binding component.
        bB : float
            Binding coefficient for weakly binding component.
        cFA : float
            Feed concentration for strongly binding component.
        cFB : float
            Feed concentration for weakly binding component.
        wG : float
            First quadratic null.
        wF : float
            Second quadratic null.
        """
        m1, m2, m3, m4 = self.calculate_m_opt(HA, HB, bA, bB, cFA, cFB, wG, wF)
        W = [m2, m3]

        R = [
            wG**2 / HA,
            wG
            * (wF * (HA - wG) * (HA - HB) + HB * wG * (HA - wF))
            / (HA * HB * (HA - wF)),
        ]

        # Bounds
        lb = W[0] - 0.3 * (HA - W[0])
        ub = HA + 0.3 * (HA - W[0])

        ax.set_xlim(lb, ub)
        ax.set_ylim(lb, ub)

        # Diagonal
        ax.plot((lb, ub), (lb, ub), "k")

        # Plot [W -> R]
        m2WR = np.linspace(W[0], R[0], 50)
        m3WR = (
            1 / (bA * cFA * wG) * (wG * (HA - wG) - (HA - wG * (1 + bA * cFA)) * m2WR)
        )
        ax.plot(m2WR, m3WR, "k-")

        # plot [W -> HB]
        m2WHB = np.linspace(W[0], HB, 10)
        m3WHB = (
            1 / (bA * cFA * HB) * (HB * (HA - HB) - (HA - HB * (1 + bA * cFA)) * m2WHB)
        )
        ax.plot(m2WHB, m3WHB, "k-")

        # plot [R -> HA]
        m2RHA = np.linspace(R[0], HA, 10)
        m3RHA = m2RHA + (np.sqrt(HA) - np.sqrt(m2RHA)) ** 2 / (bA * cFA)
        ax.plot(m2RHA, m3RHA, "k-")

        # TODO: Equations that plot regions of pure extract / raffinate not clear yet.


class CarouselSolutionBulk(SolutionBase):
    """
    Solution at unit inlet or outlet.

    N_COLUMNS * NCOL * NRAD
    """

    _coordinates = ["axial_coordinates", "radial_coordinates"]

    def __init__(
        self,
        builder: CarouselBuilder,
        simulation_results: SimulationResults,
    ) -> None:
        if not builder.column.solution_recorder.write_solution_bulk:
            raise CADETProcessError(
                "Cannot instantiate CarouselSolutionBulk if solution is not stored. "
                "Please set "
                "`builder.column.solution_recorder.write_solution_bulk = True`."
            )
        self.builder = builder
        self.simulation_results = simulation_results

    @property
    def component_system(self) -> ComponentSystem:
        return self.builder.component_system

    @property
    def solution(self) -> Dict:
        return self.simulation_results.solution

    @property
    def axial_coordinates(self) -> npt.ArrayLike:
        return self.simulation_results.solution.column_0.bulk.axial_coordinates

    @property
    def radial_coordinates(self) -> npt.ArrayLike:
        radial_coordinates = \
            self.simulation_results.solution.column_0.bulk.radial_coordinates
        if radial_coordinates is not None and len(radial_coordinates) == 1:
            radial_coordinates = None

        return radial_coordinates

    @property
    def time(self) -> float:
        return self.simulation_results.solution.column_0.bulk.time

    def plot_at_time(
        self,
        t: float,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        axs: Optional[Axes] = None,
    ) -> tuple[Figure, Axes]:
        """Plot bulk solution over space at given time.

        Parameters
        ----------
        t : float
            time for plotting
        y_min : float, optional
            Set minimum for plotting purposes.
            If None, y_min is set to minimum of the data.
        y_max : float, optional
            Set maximum for plotting purposes.
            If None, y_max is set to minimum of the data
        ax : Axes
            Axes to plot on.

        See Also
        --------
        CADETProcess.plotting
        """
        n_cols = self.builder.n_columns
        if axs is None:
            fig, axs = plt.subplots(
                ncols=n_cols,
                figsize=(n_cols * 4, 6),
                gridspec_kw=dict(wspace=0.0, hspace=0.0),
                sharey="row",
            )
        else:
            fig = axs[0].figure

        t_i = np.where(t <= self.time)[0][0]

        x = self.axial_coordinates

        y_min_data = 0
        y_max_data = 0
        zone_counter = 0
        column_counter = 0

        for position, ax in enumerate(axs):
            col_index = self.builder.column_indices_at_time(t, position)

            y = self.solution[f"column_{col_index}"].bulk.solution[t_i, :]

            y_min_data = min(y_min_data, min(0, np.min(y)))
            y_max_data = max(y_max_data, 1.1 * np.max(y))

            ax.plot(x, y)

            zone = self.builder.zones[zone_counter]

            if zone.n_columns > 1:
                ax.set_title(f"{zone.name}, position {column_counter}")
            else:
                ax.set_title(f"{zone.name}")

            if column_counter < (zone.n_columns - 1):
                column_counter += 1
            else:
                zone_counter += 1
                column_counter = 0

        plotting.add_text(ax, f"time = {t:.2f} s")

        if y_min is None:
            y_min = y_min_data
        if y_min is None:
            y_min = y_max_data

        for position, ax in enumerate(axs):
            ax.set_ylim((y_min, y_max))

        return fig, axs



from __future__ import annotations

import uuid
import warnings
from pathlib import Path
from typing import Any, Iterator, Optional

import corner
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from addict import Dict
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pymoo.visualization.scatter import Scatter

from CADETProcess import CADETProcessError, plotting
from CADETProcess.optimization.individual import Individual, hash_array


class Population:
    """
    Collection of Individuals evaluated during Optimization.

    Attributes
    ----------
    individuals : list
        Individuals evaluated during optimization.

    See Also
    --------
    CADETProcess.optimization.Individual
    ParetoFront
    """

    def __init__(self, id: Optional[str] = None) -> None:
        """
        Initialize the Population.

        Parameters
        ----------
        id : str or None, optional
            Identifier for the population. If None, a random UUID will be generated.
        """
        self._individuals = {}

        if id is None:
            self.id = uuid.uuid4()
        else:
            if isinstance(id, bytes):
                id = id.decode(encoding="utf=8")
            self.id = uuid.UUID(id)

    @property
    def feasible(self) -> "Population":
        """Population: Population containing only feasible individuals."""
        pop = Population()
        pop._individuals = {ind.id: ind for ind in self.individuals if ind.is_feasible}

        return pop

    @property
    def infeasible(self) -> "Population":
        """Population: Population containing only infeasible individuals."""
        pop = Population()
        pop._individuals = {
            ind.id: ind for ind in self.individuals if not ind.is_feasible
        }

        return pop

    @property
    def n_x(self) -> int:
        """int: Number of optimization variables."""
        return self.individuals[0].n_x

    @property
    def n_f(self) -> int:
        """int: Number of objective metrics."""
        return self.individuals[0].n_f

    @property
    def n_g(self) -> int:
        """int: Number of nonlinear constraint metrics."""
        return self.individuals[0].n_g

    @property
    def n_m(self) -> int:
        """int: Number of meta scores."""
        return self.individuals[0].n_m

    @property
    def dimensions(self) -> tuple[int]:
        """tuple: Individual dimensions (n_x, n_f, n_g, n_m)."""
        if self.n_individuals == 0:
            return None

        return self.individuals[0].dimensions

    @property
    def objectives_minimization_factors(self) -> np.ndarray:
        """np.ndarray: Array indicating objectives transformed to minimization."""
        return self.individuals[0].objectives_minimization_factors

    @property
    def meta_scores_minimization_factors(self) -> np.ndarray:
        """np.ndarray: Array indicating meta sorces transformed to minimization."""
        return self.individuals[0].meta_scores_minimization_factors

    @property
    def variable_names(self) -> list[str]:
        """list: Names of the optimization variables."""
        if self.individuals[0].variable_names is None:
            return [f"x_{i}" for i in range(self.n_x)]
        else:
            return self.individuals[0].variable_names

    @property
    def independent_variable_names(self) -> list[str]:
        """list: Names of the independent variables."""
        return self.individuals[0].independent_variable_names

    @property
    def objective_labels(self) -> list[str]:
        """list: Labels of the objective metrics."""
        return self.individuals[0].objective_labels

    @property
    def nonlinear_constraint_labels(self) -> list[str]:
        """list: Labels of the nonlinear constraint metrics."""
        return self.individuals[0].nonlinear_constraint_labels

    @property
    def meta_score_labels(self) -> list[str]:
        """list: Labels of the meta scores."""
        return self.individuals[0].meta_score_labels

    def add_individual(
        self,
        individual: Individual,
        ignore_duplicate: bool | None = True,
    ) -> None:
        """
        Add individual to population.

        Parameters
        ----------
        individual : Individual
            Individual to be added.
        ignore_duplicate : bool, optional
            If False, an Exception is thrown if the individual already exists.

        Raises
        ------
        TypeError
            If the individual is not an instance of Individual.
        CADETProcessError
            If the individual does not match the dimensions.
            If the individual already exists.
        """
        if not isinstance(individual, Individual):
            raise TypeError("Expected Individual")

        if self.dimensions is not None and individual.dimensions != self.dimensions:
            raise CADETProcessError("Individual does not match dimensions.")

        if individual in self:
            if ignore_duplicate:
                return
            else:
                raise CADETProcessError("Individual already exists.")

        self._individuals[individual.id] = individual

    def remove_individual(self, individual: Individual) -> None:
        """
        Remove an individual from the population.

        Parameters
        ----------
        individual : Individual
            Individual to be removed.

        Raises
        ------
        TypeError
            If the individual is not an instance of Individual.
        CADETProcessError
            If the individual is not in the population.
        """
        if not isinstance(individual, Individual):
            raise TypeError("Expected Individual")

        if individual not in self:
            raise CADETProcessError("Individual is not in population.")
        self._individuals.pop(individual.id)

    def update(self, other: Population) -> None:
        """
        Update the population with individuals from another population.

        Parameters
        ----------
        other : Population
            Another population.

        Raises
        ------
        TypeError
            If other is not an instance of Population.
        CADETProcessError
            If the dimensions do not match.
        """
        if not isinstance(other, Population):
            raise TypeError("Expected Population")

        if self.dimensions is not None and self.dimensions != other.dimensions:
            raise CADETProcessError("Dimensions do not match")

        self._individuals.update(other._individuals)

    def remove_similar(self) -> None:
        """Remove similar individuals from the population."""
        for ind in self.individuals.copy():
            to_remove = []

            for ind_other in self.individuals.copy():
                if ind is ind_other:
                    continue

                if ind_other.is_similar(ind, self.similarity_tol):
                    if np.any(ind_other.f == self.f_best):
                        continue
                    to_remove.append(ind_other)

            for i in reversed(to_remove):
                try:
                    self.remove_individual(i)
                except CADETProcessError:
                    pass

    @property
    def individuals(self) -> list[Individual]:
        """list: All individuals."""
        return list(self._individuals.values())

    @property
    def n_individuals(self) -> int:
        """int: Number of indivuals."""
        return len(self.individuals)

    @property
    def x(self) -> np.ndarray:
        """np.ndarray: All evaluated points."""
        return np.array([ind.x for ind in self.individuals])

    @property
    def x_transformed(self) -> np.ndarray:
        """np.ndarray: All evaluated points in independent transformed space."""
        return np.array([ind.x_transformed for ind in self.individuals])

    @property
    def cv_bounds(self) -> np.ndarray:
        """np.ndarray: All evaluated bound constraint violations."""
        return np.array([ind.cv_bounds for ind in self.individuals])

    @property
    def cv_lincon(self) -> np.ndarray:
        """np.ndarray: All evaluated linear constraint violations."""
        return np.array([ind.cv_lincon for ind in self.individuals])

    @property
    def cv_lineqcon(self) -> np.ndarray:
        """np.ndarray: All evaluated linear equality constraint violations."""
        return np.array([ind.cv_lineqcon for ind in self.individuals])

    @property
    def f(self) -> np.ndarray:
        """np.ndarray: All evaluated objective function values."""
        return np.array([ind.f for ind in self.individuals])

    @property
    def f_minimized(self) -> np.ndarray:
        """np.ndarray: All evaluated objective function values as if minimized."""
        return np.array([ind.f_min for ind in self.individuals])

    @property
    def f_best(self) -> np.ndarray:
        """np.ndarray: Best objective values."""
        f_best = np.min(self.f_minimized, axis=0)
        return np.multiply(self.objectives_minimization_factors, f_best)

    @property
    def f_min(self) -> np.ndarray:
        """np.ndarray: Minimum objective values."""
        return np.min(self.f, axis=0)

    @property
    def f_max(self) -> np.ndarray:
        """np.ndarray: Maximum objective values."""
        return np.max(self.f, axis=0)

    @property
    def f_avg(self) -> np.ndarray:
        """np.ndarray: Average objective values."""
        return np.mean(self.f, axis=0)

    @property
    def g(self) -> np.ndarray:
        """np.ndarray: All evaluated nonlinear constraint function values."""
        if self.dimensions[2] > 0:
            return np.array([ind.g for ind in self.individuals])

    @property
    def g_best(self) -> np.ndarray:
        """np.ndarray: Best nonlinear constraint values."""
        indices = np.argmin(self.cv_nonlincon, axis=0)
        return [self.g[ind, i] for i, ind in enumerate(indices)]

    @property
    def g_min(self) -> np.ndarray:
        """np.ndarray: Minimum nonlinear constraint values."""
        if self.dimensions[2] > 0:
            return np.min(self.g, axis=0)

    @property
    def g_max(self) -> np.ndarray:
        """np.ndarray: Maximum nonlinear constraint values."""
        if self.dimensions[2] > 0:
            return np.max(self.g, axis=0)

    @property
    def g_avg(self) -> np.ndarray:
        """np.ndarray: Average nonlinear constraint values."""
        if self.dimensions[2] > 0:
            return np.mean(self.g, axis=0)

    @property
    def cv_nonlincon(self) -> np.ndarray:
        """np.ndarray: All evaluated nonlinear constraint violation values."""
        if self.dimensions[2] > 0:
            return np.array([ind.cv_nonlincon for ind in self.individuals])

    @property
    def cv_nonlincon_min(self) -> np.ndarray:
        """np.ndarray: Minimum nonlinear constraint violation values."""
        if self.dimensions[2] > 0:
            return np.min(self.cv_nonlincon, axis=0)

    @property
    def cv_nonlincon_max(self) -> np.ndarray:
        """np.ndarray: Maximum nonlinearconstraint violation values."""
        if self.dimensions[2] > 0:
            return np.max(self.cv_nonlincon, axis=0)

    @property
    def cv_nonlincon_avg(self) -> np.ndarray:
        """np.ndarray: Average nonlinear constraint violation values."""
        if self.dimensions[2] > 0:
            return np.mean(self.cv_nonlincon, axis=0)

    @property
    def m(self) -> np.ndarray:
        """np.ndarray: All evaluated meta scores."""
        if self.dimensions[3] > 0:
            return np.array([ind.m for ind in self.individuals])

    @property
    def m_minimized(self) -> np.ndarray:
        """np.ndarray: All evaluated meta scores, transformed to be minimized."""
        if self.dimensions[3] > 0:
            return np.array([ind.m_min for ind in self.individuals])

    @property
    def m_best(self) -> np.ndarray:
        """np.ndarray: Best meta scores."""
        if self.dimensions[3] > 0:
            m_best = np.min(self.m_minimized, axis=0)
            return np.multiply(self.meta_scores_minimization_factors, m_best)

    @property
    def m_min(self) -> np.ndarray:
        """np.ndarray: Minimum meta scores."""
        if self.dimensions[3] > 0:
            return np.min(self.m, axis=0)

    @property
    def m_max(self) -> np.ndarray:
        """np.ndarray: Maximum meta scores."""
        if self.dimensions[3] > 0:
            return np.max(self.m, axis=0)

    @property
    def m_avg(self) -> np.ndarray:
        """np.ndarray: Average meta scores."""
        if self.dimensions[3] > 0:
            return np.mean(self.m, axis=0)

    @property
    def is_feasilbe(self) -> bool:
        """np.ndarray: False if any constraint is not met. True otherwise."""
        return np.array([ind.is_feasible for ind in self.individuals])

    def setup_objectives_figure(
        self,
        include_meta: Optional[bool] = True,
        plot_individual: Optional[bool] = False,
    ) -> tuple:
        """
        Set up figure and axes for plotting objectives.

        Parameters
        ----------
        include_meta : bool, optional
            If True, include meta scores in the plot. The default is True.
        plot_individual : bool, optional
            If True, create separate figures for each objective.
            Otherwise, plot all objectives in one figure. The default is True.

        Returns
        -------
        tuple
            A tuple of the figure(s) and axes object(s).
        """
        n = len(self.variable_names)
        if include_meta and self.m is not None:
            m = len(self.objective_labels) + len(self.meta_score_labels)
        else:
            m = len(self.objective_labels)

        if n == 0:
            return (None, None)

        space_fig_all, space_axs_all = plt.subplots(
            nrows=m,
            ncols=n,
            figsize=(n * 8 + 2, m * 8 + 2),
            squeeze=False,
        )
        plt.close(space_fig_all)

        space_figs_ind = []
        space_axs_ind = []
        for i in range(m * n):
            fig, ax = plt.subplots()
            space_figs_ind.append(fig)
            space_axs_ind.append(ax)
            plt.close(fig)

        space_axs_ind = np.array(space_axs_ind).reshape(space_axs_all.shape)

        if plot_individual:
            return space_figs_ind, space_axs_ind
        else:
            return space_fig_all, space_axs_all

    def plot_objectives(
        self,
        figs: Optional[Figure | list[Figure]] = None,
        axs: Optional[Axes | list[list[Axes]]] = None,
        include_meta: bool = True,
        plot_infeasible: bool = True,
        plot_individual: bool = False,
        autoscale: bool = True,
        color_feas: str = "blue",
        color_infeas: str = "red",
        show: bool = True,
        plot_directory: Optional[str | Path] = None,
    ) -> tuple[list[Figure], list[list[Axes]]]:
        """
        Plot the objective function values for each design variable.

        Parameters
        ----------
        figs : plt.Figure or list, optional
            Figure(s) to plot the objectives on. The default is None.
        axs : plt.Axes or list, optional
            Axes to plot the objectives on. The default is None.
        include_meta : bool, optional
            If True, include meta scores in the plot. The default is True.
        plot_infeasible : bool, optional
            If True, plot infeasible points. The default is True.
        plot_individual : bool, optional
            If True, create separate figures for each objective.
            Otherwise, plot all objectives in one figure. The default is False.
        autoscale : bool, optional
            If True, automatically adjust the scaling of the axes. The default is True.
        color_feas : str, optional
            The color for the feasible points. The default is 'blue'.
        color_infeas : str, optional
            The color for the infeasible points. The default is 'red'.
        show : bool, optional
            If True, display the plot. The default is True.
        plot_directory : str, optional
            The directory where the plot should be saved. The default is None.

        Returns
        -------
        tuple
            A tuple of the figure(s) and axes object(s).
        """
        if axs is None:
            figs, axs = self.setup_objectives_figure(include_meta, plot_individual)

        if not isinstance(figs, list):
            figs = [figs]

        layout = plotting.Layout()
        layout.y_label = "$f~/~-$"

        variables = self.variable_names
        feasible = self.feasible
        infeasible = self.infeasible
        x_feas = feasible.x
        x_infeas = infeasible.x

        if include_meta and self.m is not None:
            if len(feasible) > 0:
                values_feas = np.hstack((feasible.f, feasible.m))
            else:
                values_infeas = np.empty((0, self.n_f + self.n_m))
            if len(infeasible) > 0:
                values_infeas = np.hstack((infeasible.f, infeasible.m))
            else:
                values_infeas = np.empty((0, self.n_f + self.n_m))

            labels = self.objective_labels + self.meta_score_labels
        else:
            values_feas = feasible.f
            values_infeas = infeasible.f
            labels = self.objective_labels

        for i_var, var in enumerate(variables):
            if len(feasible) > 0:
                x_var_feas = x_feas[:, i_var]
            if len(infeasible) > 0:
                x_var_infeas = x_infeas[:, i_var]

            for i_metric, label in enumerate(labels):
                ax = axs[i_metric][i_var]

                if len(feasible) > 0:
                    v_metric_feas = values_feas[:, i_metric]
                    ax.scatter(x_var_feas, v_metric_feas, alpha=0.5, color=color_feas)

                if len(infeasible) > 0 and plot_infeasible:
                    v_metric_infeas = values_infeas[:, i_metric]
                    ax.scatter(
                        x_var_infeas, v_metric_infeas, alpha=0.5, color=color_infeas
                    )

                points = np.vstack([col.get_offsets() for col in ax.collections])

                x_all = points[:, 0]
                v_all = points[:, 1]

                layout.x_lim = (np.nanmin(x_all), np.nanmax(x_all))
                layout.x_label = var
                if autoscale and np.min(x_all) > 0:
                    if np.max(x_all) / np.min(x_all[x_all > 0]) > 100.0:
                        ax.set_xscale("log")

                y_min = np.nanmin(v_all)
                y_max = np.nanmax(v_all)
                y_lim = (min(0.9 * y_min, y_min - 0.01 * (y_max - y_min)), 1.1 * y_max)
                layout.y_label = label
                if autoscale and np.min(v_all) > 0:
                    if np.max(v_all) / np.min(v_all[v_all > 0]) > 100.0:
                        ax.set_yscale("log")
                        y_lim = (y_min / 2, y_max * 2)
                if y_min != y_max:
                    layout.y_lim = y_lim

                try:
                    plotting.set_layout(ax, layout)
                except ValueError:
                    pass

        for fig in figs:
            fig.tight_layout()
            if not show:
                plt.close(fig)
            else:
                dummy = plt.figure(figsize=fig.get_size_inches())
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                plt.show()

        if plot_directory is not None:
            plot_directory = Path(plot_directory)
            if plot_individual:
                for i, fig in enumerate(figs):
                    fig.savefig(f"{plot_directory / 'objectives'}_{i}.png")
            else:
                figs[0].savefig(f"{plot_directory / 'objectives'}.png")

        if plot_individual:
            return figs, axs
        else:
            return figs[0], axs

    def setup_pareto(self, include_meta: bool = False) -> Scatter:
        """
        Set up base figure for plotting the Pareto front.

        Parameters
        ----------
        include_meta : bool
            If True, include meta scores in Pareto plot.

        Returns
        -------
        pymoo.visualization.scatter.Scatter
            The base figure object.
        """
        if include_meta:
            n = self.dimensions[1] + self.dimensions[3]
            labels = self.objective_labels + self.meta_score_labels
        else:
            n = self.dimensions[1]
            labels = self.objective_labels
        plot = Scatter(
            figsize=(6 * n, 5 * n),
            tight_layout=True,
            plot_3d=False,
            labels=labels,
        )
        return plot

    def plot_pareto(
        self,
        plot: Optional[Scatter] = None,
        include_meta: bool = True,
        plot_infeasible: bool = True,
        color_feas: str = "blue",
        color_infeas: str = "red",
        show: bool = True,
        plot_directory: Optional[str | Path] = None,
    ) -> Scatter:
        """
        Plot pairwise Pareto fronts for each generation in the optimization.

        The Pareto front represents the optimal solutions that cannot be improved in one
        objective without sacrificing another. The method shows a pairwise Pareto plot,
        where each objective is plotted against every other objective in a scatter plot,
        allowing for a visualization of the trade-offs between the objectives.

        Parameters
        ----------
        plot : pymoo.visualization.scatter.Scatter, optional
            Base figure. If None is provided, a new one will be set up.
        include_meta : bool, optional
            If True, include meta scores in the plot. The default is True.
        plot_infeasible : bool, optional
            If True, plot infeasible points. The default is True.
        color_feas : str, optional
            The color for the feasible points. The default is 'blue'.
        color_infeas : str, optional
            The color for the infeasible points. The default is 'red'.
        show : bool, optional
            If True, display the plot. The default is True.
        plot_directory : str, optional
            The directory where the plot should be saved. The default is None.

        Returns
        -------
        pymoo.visualization.scatter.Scatter
            The scatter plot object.
        """
        if plot is None:
            plot = self.setup_pareto(include_meta)

        feasible = self.feasible
        infeasible = self.infeasible

        if include_meta and self.m is not None:
            if len(feasible) > 0:
                values_feas = np.hstack((feasible.f, feasible.m))
            else:
                values_infeas = np.empty((0, self.n_f + self.n_m))
            if len(infeasible) > 0:
                values_infeas = np.hstack((infeasible.f, infeasible.m))
            else:
                values_infeas = np.empty((0, self.n_f + self.n_m))
        else:
            values_feas = feasible.f
            values_infeas = infeasible.f

        if len(feasible) > 0:
            plot.add(values_feas, s=10, color=color_feas)

        if plot_infeasible and len(infeasible) > 0:
            plot.add(values_infeas, s=10, color=color_infeas)

        if plot_directory is not None:
            plot_directory = Path(plot_directory)
            plot.save(f"{plot_directory / 'pareto.png'}")

        if not show:
            plt.close(plot.fig)
        else:
            plot.show()

        return plot

    def plot_pairwise(
        self,
        fig: Optional[plt.Figure] = None,
        axs: Optional[npt.NDArray[plt.Axes]] = None,
        n_bins: int = 20,
        use_transformed: bool = False,
        autoscale: bool = True,
        show: bool = True,
        plot_directory: Optional[str] = None,
    ) -> tuple[plt.Figure, np.ndarray]:
        """
        Create a pairplot using Matplotlib.

        Parameters
        ----------
        fig : Optional[plt.Figure], default=None
            An optional Matplotlib Figure object. If none is provided, a new figure will
            be created.
        axs : Optional[npt.NDArray[plt.Axes]], default=None
            An optional array of Matplotlib Axes. If none is provided, new axes will
            be created.
        n_bins : int, default=20
            Number of bins for histogram plots.
        use_transformed : bool, optional
            If True, use the transformed independent variables. The default is False.
        autoscale : bool, optional
            If True, automatically adjust the scaling of the axes. The default is True.
        use_transformed : bool, optional
            If True, transformed values will be plotted. The default is False.
        show : bool, optional
            If True, display the plot. The default is True.
        plot_directory : str, optional
            The directory where the plot should be saved. The default is None.

        Returns
        -------
        tuple
            A tuple containing:
            - plt.Figure: The Matplotlib Figure object.
            - np.ndarray: An array of Axes objects representing the subplot grid.
        """
        if use_transformed:
            x = self.x_transformed
            labels = self.independent_variable_names
        else:
            x = self.x
            labels = self.variable_names

        fig, axs = plot_pairwise(
            x,
            labels,
            n_bins=n_bins,
            autoscale=autoscale,
            fig=fig,
            axs=axs,
        )

        if plot_directory is not None:
            plot_directory = Path(plot_directory)
            fig.savefig(f"{plot_directory / 'pairwise.png'}")

        if not show:
            plt.close(fig)

        return fig, axs

    def plot_corner(
        self,
        use_transformed: bool = False,
        show: bool = True,
        plot_directory: Optional[str] = None,
    ) -> None:
        """
        Create a corner plot of the independent variables.

        Parameters
        ----------
        use_transformed : bool, optional
            If True, use the transformed independent variables. The default is False.
        show : bool, optional
            If True, display the plot. The default is True.
        plot_directory : str, optional
            The directory where the plot should be saved. The default is None.
        """
        warnings.warn(
            "This method will be deprecated in the future. "
            "Use `plot_pairwise` instead.",
            FutureWarning,
        )

        if use_transformed:
            x = self.x_transformed
            labels = self.independent_variable_names
        else:
            x = self.x
            labels = self.variable_names

        # To avoid error, remove dimensions where all entries are the same value.
        singular_indices = []
        singular_labels = []

        for i, col in enumerate(x.transpose()):
            if len(np.unique(col)) == 1:
                singular_indices.append(i)
                singular_labels.append(labels[i])

        x = np.delete(x.transpose(), singular_indices, 0).transpose()
        labels = [label for label in labels if label not in singular_labels]

        fig = corner.corner(
            x,
            labels=labels,
            bins=20,
            quantiles=[0.16, 0.5, 0.84],
            show_titles=True,
            title_kwargs={"fontsize": 20},
            title_fmt=".2g",
            use_math_text=True,
            quiet=True,
        )
        fig_size = 6 * len(labels)
        fig.set_size_inches((fig_size, fig_size))
        fig.tight_layout()

        if plot_directory is not None:
            plot_directory = Path(plot_directory)
            fig.savefig(f"{plot_directory / 'corner.png'}")

        if not show:
            plt.close(fig)

    def __contains__(self, other: Individual | np.ndarray | list) -> bool:
        """
        Check if the population contains a specific individual.

        Parameters
        ----------
        other : Individual | np.ndarray | list
            The individual or its hashable representation.

        Returns
        -------
        bool
            True if the individual is in the population, False otherwise.
        """
        if isinstance(other, Individual):
            key = other.id
        elif isinstance(other, (np.array, list)):
            key = hash_array(other)
        else:
            key = None

        if key in self._individuals:
            return True
        else:
            return False

    def __getitem__(self, x: np.ndarray | list) -> Individual:
        """
        Get an individual from the population using its hashable representation.

        Parameters
        ----------
        x : np.ndarray | list
            The hashable representation of the individual.

        Returns
        -------
        Individual
            The individual from the population.
        """
        key = hash_array(x)

        return self._individuals[key]

    def __len__(self) -> int:
        """
        Get the number of individuals in the population.

        Returns
        -------
        int
            The number of individuals in the population.
        """
        return self.n_individuals

    def __iter__(self) -> Iterator[Individual]:
        """
        Iterate over the individuals in the population.

        Returns
        -------
        iter
            An iterator over the individuals in the population.
        """
        return iter(self.individuals)

    def to_dict(self) -> Dict:
        """
        Convert Population to a dictionary.

        Returns
        -------
        dict
            Population as a dictionary with individuals stored as list of dictionaries.
        """
        data = Dict()
        data.id = str(self.id)

        for i, ind in enumerate(self.individuals):
            data.individuals[i] = ind.to_dict()

        return data

    @classmethod
    def from_dict(cls, data: dict) -> Population:
        """
        Create a Population from a dictionary.

        Parameters
        ----------
        data : dict
            The dictionary containing population data.

        Returns
        -------
        Population
            The Population created from the data.
        """
        id = data["id"]
        if isinstance(id, bytes):
            id = id.decode(encoding="utf=8")
        population = cls(id)
        for individual_data in data["individuals"].values():
            individual = Individual.from_dict(individual_data)
            population.add_individual(individual)
        return population


class ParetoFront(Population):
    """Class representing a Pareto front in a multi-objective optimization problem."""

    def __init__(
        self,
        similarity_tol: float = 1e-1,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a ParetoFront with a specified similarity tolerance.

        Parameters
        ----------
        similarity_tol : float, optional
            Tolerance for similarity between individuals. Default is 1e-1.
        *args : tuple
            Additional positional arguments for the parent class.
        **kwargs : dict
            Additional keyword arguments for the parent class.
        """
        self.similarity_tol = similarity_tol
        super().__init__(*args, **kwargs)

    def update_population(self, population: Population) -> tuple[list, bool]:
        """
        Update the Pareto front with a new population.

        Parameters
        ----------
        population : Population
            The population used to update the Pareto front.

        Returns
        -------
        tuple[list, bool]
            A tuple containing new members added to the Pareto front and a boolean indicating
            if there was a significant improvement.
        """
        new_members = []
        significant = []

        for ind_new in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []

            if not ind_new.is_feasible:
                continue

            for i, ind_pareto in enumerate(self):
                # Do not add if is dominated
                if not dominates_one and ind_pareto.dominates(ind_new):
                    is_dominated = True
                    break

                # Remove existing if infeasible
                elif not ind_pareto.is_feasible:
                    dominates_one = True
                    to_remove.append(ind_pareto)
                    significant.append(True)

                # Remove existing if new dominates
                elif ind_new.dominates(ind_pareto):
                    dominates_one = True
                    to_remove.append(ind_pareto)
                    if not ind_new.is_similar(ind_pareto, self.similarity_tol):
                        significant.append(True)

                # Ignore similar individuals
                elif ind_new.is_similar(ind_pareto, self.similarity_tol):
                    has_twin = True
                    break

            for i in reversed(to_remove):
                self.remove_individual(i)

            if not is_dominated:
                if len(self) == 0:
                    significant.append(True)
                if not has_twin:
                    significant.append(True)

                self.add_individual(ind_new)
                new_members.append(ind_new)

        if len(self) == 0:
            # Use least inveasible individuals.
            indices = np.argmin(population.cv_bounds, axis=0)
            for index in indices:
                ind_new = population.individuals[index]
                self.add_individual(ind_new)

            indices = np.argmin(population.cv_lincon, axis=0)
            for index in indices:
                ind_new = population.individuals[index]
                self.add_individual(ind_new)

            indices = np.argmin(population.cv_lineqcon, axis=0)
            for index in indices:
                ind_new = population.individuals[index]
                self.add_individual(ind_new)

            if self.n_g > 0:
                indices = np.argmin(population.cv_nonlincon, axis=0)
                for index in indices:
                    ind_new = population.individuals[index]
                    self.add_individual(ind_new)

        elif len(self) > 1:
            self.remove_infeasible()

        if self.similarity_tol:
            self.remove_similar()

        return new_members, any(significant)

    def remove_infeasible(self) -> None:
        """Remove infeasible individuals from the Pareto front."""
        for ind in self.individuals.copy():
            if not ind.is_feasible:
                self.remove_individual(ind)

    def remove_dominated(self) -> None:
        """Remove dominated individuals from the Pareto front."""
        for ind in self.individuals.copy():
            dominates_one = False
            to_remove = []

            for ind_other in self.individuals.copy():
                if not dominates_one and ind_other.dominates(ind):
                    to_remove.append(ind)
                    break
                elif ind.dominates(ind_other):
                    dominates_one = True
                    to_remove.append(ind_other)

            for i in reversed(to_remove):
                try:
                    self.remove_individual(i)
                except CADETProcessError:
                    pass

    def to_dict(self) -> dict:
        """
        Convert the ParetoFront to a dictionary.

        Returns
        -------
        dict
            A dictionary representation of the ParetoFront, including individuals and
            similarity tolerance if set.
        """
        front = super().to_dict()
        if self.similarity_tol:
            front["similarity_tol"] = self.similarity_tol

        return front

    @classmethod
    def from_dict(cls, data: dict) -> ParetoFront:
        """
        Create a ParetoFront instance from a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary containing the ParetoFront data.

        Returns
        -------
        ParetoFront
            An instance of ParetoFront created from the dictionary.
        """
        front = cls(similarity_tol=data.get("similarity_tol"), id=data["id"])
        for individual_data in data["individuals"].values():
            individual = Individual.from_dict(individual_data)
            front.add_individual(individual)

        return front


def plot_pairwise(
    population: npt.ArrayLike,
    variable_names: Optional[list[str]] = None,
    n_bins: int = 20,
    autoscale: bool = True,
    fig: Optional[plt.Figure] = None,
    axs: Optional[np.ndarray[plt.Axes]] = None,
) -> tuple[plt.Figure, np.ndarray[plt.Axes]]:
    """
    Create a pairwise scatter plot for all variables of a population.

    Parameters
    ----------
    population : npt.ArrayLike
        3D array-like structure containing numerical variables with shape
        (n_chains, n_samples, n_variables)
    variable_names : list of str, optional
        list of variable names corresponding to columns in the data.
        If None, default names will be assigned.
    n_bins : int, default=20
        Number of bins for histogram plots.
    autoscale : bool, default=True
        If True, automatically adjust the scaling of the axes.
    fig : Optional[plt.Figure], default=None
        An optional Matplotlib Figure object. If none is provided, a new figure will be
        created.
    axs : Optional[npt.NDArray[plt.Axes]], default=None
        An optional array of Matplotlib Axes. If none is provided, new axes will be
        created.

    Returns
    -------
    tuple
        A tuple containing:
        - plt.Figure: The Matplotlib Figure object.
        - npt.NDArray[plt.Axes]: An array of Axes objects representing the subplot grid.
    """
    population = np.array(population)

    if population.ndim != 2:
        raise ValueError(f"Expected 2D array, got array with ndim={population.ndim}")

    n_samples, n_variables = population.shape

    if variable_names is None:
        variable_names = [f"$x_{i}$" for i in range(n_variables)]

    if fig is None and axs is None:
        fig, axs = plt.subplots(
            n_variables,
            n_variables,
            figsize=(6 * n_variables, 5 * n_variables),
            sharex="col",
            sharey="row",
            squeeze=False,
        )

    if axs.shape != (n_variables, n_variables):
        raise ValueError(
            "Inconsistent shape for provided axes."
            f"Expected {(n_variables, n_variables)}, got {axs.shape}."
        )

    # Rows
    for i in range(n_variables):
        scale_i = False
        if autoscale and np.all(population[:, i] > 0):
            value_range = population[:, i].max() / population[:, i].min()
            if value_range > 100.0:
                scale_i = True

        # Columns
        for j in range(n_variables):
            scale_j = False
            if autoscale and np.all(population[:, j] > 0):
                value_range = population[:, j].max() / population[:, j].min()
                if value_range > 100.0:
                    scale_j = True

            ax = axs[i, j]
            if i == j:
                # Create a twin axis for histograms to avoid sharing y-axis
                ax_hist = ax.twinx()

                # Determine binning strategy
                if scale_i:
                    bins = np.geomspace(
                        population[:, i].min(), population[:, i].max(), n_bins + 1
                    )
                else:
                    bins = np.linspace(
                        population[:, i].min(), population[:, i].max(), n_bins + 1
                    )

                ax_hist.hist(
                    population[:, i],
                    bins=bins,
                    alpha=0.7,
                    color="blue",
                    edgecolor="black",
                    align="mid",
                )
                ax_hist.set_yticks([])  # Hide y-ticks for the histogram
            else:
                # Scatter plot for non-diagonal elements
                ax.scatter(population[:, j], population[:, i], alpha=0.5, s=10)

            # Apply log scale based on autoscale logic
            if scale_j:
                ax.set_xscale("log")
            if scale_i:
                ax.set_yscale("log")

            # Ensure axis labels and ticks are visible only on the
            # first column
            if j == 0:
                ax.yaxis.set_tick_params(labelleft=True)
                if not scale_i:
                    ax.ticklabel_format(axis="y", useMathText=True, scilimits=[-3, 3])
            else:
                ax.yaxis.set_tick_params(labelleft=False)

            # and last row
            if i == n_variables - 1:
                ax.xaxis.set_tick_params(labelbottom=True)
                if not scale_j:
                    ax.ticklabel_format(axis="x", useMathText=True, scilimits=[-3, 3])
            else:
                ax.xaxis.set_tick_params(labelbottom=False)

            # Set axis labels on the edges
            if i == n_variables - 1:
                ax.set_xlabel(variable_names[j])
            if j == 0:
                ax.set_ylabel(variable_names[i])

    fig.tight_layout()

    return fig, axs



import shutil
import unittest
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.testing as npt
import pytest
from CADETProcess import CADETProcessError, SimulationResults
from CADETProcess.processModel import Process
from CADETProcess.processModel.discretization import NoDiscretization
from CADETProcess.simulator import Cadet

from tests.create_LWE import create_lwe


def detect_cadet(install_path: Optional[Path] = None):
    try:
        simulator = Cadet(install_path)
        found_cadet = True
        install_path = simulator.install_path
    except FileNotFoundError:
        found_cadet = False

    return found_cadet, install_path


install_path = None
found_cadet, install_path = detect_cadet()


class Test_Adapter(unittest.TestCase):
    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_check_cadet(self):
        simulator = Cadet(install_path)

        self.assertTrue(simulator.check_cadet())

        file_name = simulator.get_tempfile_name()
        cwd = simulator.temp_dir
        sim = simulator.get_new_cadet_instance()
        sim.create_lwe(cwd / file_name)

        # Check for CADET-Python >= v1.1, which introduced the .run_simulation interface.
        # If it's not present, assume CADET-Python <= 1.0.4 and use the old .run_load() interface
        # This check can be removed at some point in the future.
        if hasattr(sim, "run_simulation"):
            return_information = sim.run_simulation()
        else:
            return_information = sim.run_load()

    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_create_lwe(self):
        simulator = Cadet(install_path)

        file_name = simulator.get_tempfile_name()
        cwd = simulator.temp_dir
        sim = simulator.get_new_cadet_instance()
        sim.create_lwe(cwd / file_name)

        # Check for CADET-Python >= v1.1, which introduced the .run_simulation interface.
        # If it's not present, assume CADET-Python <= 1.0.4 and use the old .run_load() interface
        # This check can be removed at some point in the future.
        if hasattr(sim, "run_simulation"):
            return_information = sim.run_simulation()
        else:
            return_information = sim.run_load()

    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_version(self):
        simulator = Cadet(install_path)
        version_pattern = r"\d\.\d\.\d"
        self.assertRegex(
            simulator.version, version_pattern, "Version format should be X.X.X"
        )

    def tearDown(self):
        shutil.rmtree("./tmp", ignore_errors=True)


unit_types = [
    "Cstr",
    "GeneralRateModel",
    "TubularReactor",
    "LumpedRateModelWithoutPores",
    "LumpedRateModelWithPores",
    "MCT",
]

parameter_combinations = [
    {},  # Default parameters
    {"n_par": 1},
    {"n_col": 1},
]

# Parameters to skip for specific unit types
exclude_rules = {
    "Cstr": [{"n_col": 1}, {"n_par": 1}],
    "TubularReactor": [{"n_par": 1}],
    "LumpedRateModelWithoutPores": [{"n_par": 1}],
    "LumpedRateModelWithPores": [{"n_par": 1}],
    "MCT": [{"n_par": 1}],
}


# Helper function to format parameters
def format_params(params):
    if not params:
        return "default"
    return "-".join(f"{k}={v}" for k, v in params.items())


process_test_cases = [
    pytest.param(
        (unit_type, params),
        id=f"{unit_type}-{format_params(params)}",
    )
    for unit_type in unit_types
    for params in parameter_combinations
    if not (unit_type in exclude_rules and params in exclude_rules[unit_type])
]

use_dll_options = [True, False]

simulation_test_cases = [
    pytest.param(
        (unit_type, params, use_dll),
        id=f"{unit_type}-{format_params(params)}-dll={use_dll}",
    )
    for unit_type in unit_types
    for params in parameter_combinations
    for use_dll in use_dll_options
    if not (unit_type in exclude_rules and params in exclude_rules[unit_type])
]


def run_simulation(
    process: Process, install_path: Optional[str] = None, use_dll: bool = False
) -> SimulationResults:
    """
    Run the CADET simulation for the given process and handle potential issues.

    Parameters
    ----------
    process : Process
        The process to simulate.
    install_path : str, optional
        The path to the CADET installation.

    Returns
    -------
    SimulationResults
        The results of the simulation.

    Raises
    ------
    CADETProcessError
        If the simulation fails with an error.
    """
    try:
        process_simulator = Cadet(install_path)
        process_simulator.use_dll = use_dll
        simulation_results = process_simulator.simulate(process)

        if not simulation_results.exit_flag == 0:
            raise CADETProcessError(
                f"LWE simulation failed with {simulation_results.exit_message}."
            )

        return simulation_results

    except Exception as e:
        raise CADETProcessError(f"CADET simulation failed: {e}.") from e


@pytest.fixture()
def process(request: pytest.FixtureRequest):
    """
    Fixture to set up the process for each unit type without running the simulation.
    """
    unit_type, kwargs = request.param
    process = create_lwe(unit_type, **kwargs)
    return process


@pytest.fixture
def simulation_results(request: pytest.FixtureRequest):
    """
    Fixture to set up the simulation for each unit type with different `use_dll` options.
    """
    unit_type, kwargs, use_dll = request.param  # Extract `use_dll`
    process = create_lwe(unit_type, **kwargs)  # Process remains unchanged
    simulation_results = run_simulation(process, install_path, use_dll=use_dll)
    return simulation_results


@pytest.mark.parametrize("process", process_test_cases, indirect=True)
@pytest.mark.slow
class TestProcessWithLWE:
    def return_process_config(self, process: Process) -> dict:
        """
        Returns the process configuration.

        Parameters
        ----------
        process : Process
            The process object.

        Returns
        -------
        dict
            The configuration of the process.
        """
        process_simulator = Cadet(install_path)
        process_config = process_simulator.get_process_config(process).input
        return process_config

    def test_model_config(self, process: Process):
        """
        Test the model configuration for various unit types in the process.

        Parameters
        ----------
        process : Process
            The process object.
        """
        process_config = self.return_process_config(process)

        n_comp = process.component_system.n_comp
        unit = process.flow_sheet.units[1]

        model_config = process_config.model
        input_config = model_config.unit_000
        output_config = model_config.unit_002
        unit_config = model_config.unit_001

        # ASSERT INPUT CONFIGURATION
        c1_lwe = [[50.0], [0.0], [[100.0, 0.2]]]
        cx_lwe = [[1.0], [0.0], [0.0]]

        expected_input_config = {
            "UNIT_TYPE": "INLET",
            "NCOMP": n_comp,
            "INLET_TYPE": "PIECEWISE_CUBIC_POLY",
            "discretization": {"nbound": n_comp * [0]},
            "sec_000": {
                "const_coeff": np.array(c1_lwe[0] + cx_lwe[0] * (n_comp - 1)),
                "lin_coeff": np.array([0.0] * n_comp),
                "quad_coeff": np.array([0.0] * n_comp),
                "cube_coeff": np.array([0.0] * n_comp),
            },
            "sec_001": {
                "const_coeff": np.array(c1_lwe[1] + cx_lwe[1] * (n_comp - 1)),
                "lin_coeff": np.array([0.0] * n_comp),
                "quad_coeff": np.array([0.0] * n_comp),
                "cube_coeff": np.array([0.0] * n_comp),
            },
            "sec_002": {
                "const_coeff": np.array([c1_lwe[2][0][0]] + cx_lwe[2] * (n_comp - 1)),
                "lin_coeff": np.array([c1_lwe[2][0][1]] + cx_lwe[2] * (n_comp - 1)),
                "quad_coeff": np.array([0.0] * n_comp),
                "cube_coeff": np.array([0.0] * n_comp),
            },
        }

        npt.assert_equal(input_config, expected_input_config)

        # ASSERT OUTPUT CONFIGURATION
        expected_output_config = {
            "UNIT_TYPE": "OUTLET",
            "NCOMP": n_comp,
            "discretization": {"nbound": n_comp * [0]},
        }

        npt.assert_equal(output_config, expected_output_config)

        # ASSERT MODEL CONFIGURATION
        assert unit_config.NCOMP == n_comp
        assert model_config.nunits == 3

        if unit.name == "Cstr":
            self.check_cstr(unit, unit_config)
        elif unit.name == "GeneralRateModel":
            self.check_general_rate_model(unit, unit_config)
        elif unit.name == "TubularReactor":
            self.check_tubular_reactor(unit, unit_config)
        elif unit.name == "LumpedRateModelWithoutPores":
            self.check_lumped_rate_model_without_pores(unit, unit_config)
        elif unit.name == "LumpedRateModelWithPores":
            self.check_lumped_rate_model_with_pores(unit, unit_config)
        elif unit.name == "MCT":
            self.check_mct(unit, unit_config)

    def check_cstr(self, unit, unit_config):
        """
        Check the configuration for a CSTR unit.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp

        assert unit_config.UNIT_TYPE == "CSTR"
        assert unit_config.INIT_Q == n_comp * [0]
        assert unit_config.INIT_C == n_comp * [0]
        assert unit_config.INIT_LIQUID_VOLUME == 0.0008425
        assert unit_config.CONST_SOLID_VOLUME == 0.00015749999999999998
        assert unit_config.FLOWRATE_FILTER == 0.0
        assert unit_config.nbound == [1, 1, 1, 1]

        self.check_adsorption_config(unit, unit_config)

    def check_general_rate_model(self, unit, unit_config):
        """
        Check the configuration for a General Rate Model unit.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp

        assert unit_config.UNIT_TYPE == "GENERAL_RATE_MODEL"
        assert unit_config.INIT_Q == n_comp * [0]
        assert unit_config.INIT_C == n_comp * [0]
        assert unit_config.INIT_CP == n_comp * [0]
        assert unit_config.VELOCITY == unit.flow_direction
        assert unit_config.COL_DISPERSION == n_comp * [5.75e-08]
        assert unit_config.CROSS_SECTION_AREA == np.pi * 0.01**2
        assert unit_config.COL_LENGTH == 0.014
        assert unit_config.COL_POROSITY == 0.37
        assert unit_config.FILM_DIFFUSION == [6.9e-6] * n_comp

        self.check_particle_config(unit_config)
        self.check_adsorption_config(unit, unit_config)
        self.check_discretization(unit, unit_config)

    def check_tubular_reactor(self, unit, unit_config):
        """
        Check the configuration for a Tubular Reactor unit.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp

        assert unit_config.UNIT_TYPE == "LUMPED_RATE_MODEL_WITHOUT_PORES"
        assert unit_config.INIT_C == n_comp * [0]
        assert unit_config.VELOCITY == unit.flow_direction
        assert unit_config.COL_DISPERSION == n_comp * [5.75e-08]
        assert unit_config.CROSS_SECTION_AREA == np.pi * 0.01**2
        assert unit_config.COL_LENGTH == 0.014
        assert unit_config.TOTAL_POROSITY == 1

        self.check_discretization(unit, unit_config)

    def check_lumped_rate_model_without_pores(self, unit, unit_config):
        """
        Check the configuration for a Lumped Rate Model Without Pores unit.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp

        assert unit_config.UNIT_TYPE == "LUMPED_RATE_MODEL_WITHOUT_PORES"
        assert unit_config.INIT_C == n_comp * [0]
        assert unit_config.VELOCITY == unit.flow_direction
        assert unit_config.COL_DISPERSION == n_comp * [5.75e-08]
        assert unit_config.CROSS_SECTION_AREA == np.pi * 0.01**2
        assert unit_config.COL_LENGTH == 0.014
        assert unit_config.TOTAL_POROSITY == 0.8425

        self.check_adsorption_config(unit, unit_config)
        self.check_discretization(unit, unit_config)

    def check_lumped_rate_model_with_pores(self, unit, unit_config):
        """
        Check the configuration for a Lumped Rate Model With Pores unit.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp

        assert unit_config.UNIT_TYPE == "LUMPED_RATE_MODEL_WITH_PORES"
        assert unit_config.INIT_C == n_comp * [0]
        assert unit_config.INIT_Q == n_comp * [0]
        assert unit_config.INIT_CP == n_comp * [0]
        assert unit_config.VELOCITY == unit.flow_direction
        assert unit_config.COL_DISPERSION == n_comp * [5.75e-08]
        assert unit_config.CROSS_SECTION_AREA == np.pi * 0.01**2
        assert unit_config.COL_LENGTH == 0.014
        assert unit_config.COL_POROSITY == 0.37
        assert unit_config.FILM_DIFFUSION == [6.9e-6] * n_comp

        self.check_adsorption_config(unit, unit_config)
        self.check_discretization(unit, unit_config)
        self.check_particle_config(unit_config)

    def check_mct(self, unit, unit_config):
        """
        Check the configuration for a Multi-Channel Transport unit.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp
        n_channel = unit.nchannel

        assert unit_config.UNIT_TYPE == "MULTI_CHANNEL_TRANSPORT"
        npt.assert_equal(unit_config.INIT_C, n_comp * [(n_channel * [0])])
        assert unit_config.VELOCITY == unit.flow_direction
        npt.assert_equal(
            unit_config.EXCHANGE_MATRIX,
            np.array(
                [
                    [n_comp * [0.0], n_comp * [0.001], n_comp * [0.0]],
                    [n_comp * [0.002], n_comp * [0.0], n_comp * [0.003]],
                    [n_comp * [0.0], n_comp * [0.0], n_comp * [0.0]],
                ]
            ),
        )
        assert unit_config.COL_DISPERSION == n_comp * n_channel * [5.75e-08]
        assert unit_config.NCHANNEL == 3
        assert unit_config.CHANNEL_CROSS_SECTION_AREAS == 3 * [2 * np.pi * (0.01**2)]

        self.check_discretization(unit, unit_config)

    def check_adsorption_config(self, unit, unit_config):
        """
        Check the adsorption configuration.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        n_comp = unit.component_system.n_comp

        expected_adsorption_config = {
            "ADSORPTION_MODEL": "STERIC_MASS_ACTION",
            "IS_KINETIC": True,
            "SMA_KA": n_comp * [35.5],
            "SMA_KD": n_comp * [1000.0],
            "SMA_LAMBDA": 1200.0,
            "SMA_NU": n_comp * [4.7],
            "SMA_SIGMA": n_comp * [11.83],
            "SMA_REFC0": 1.0,
            "SMA_REFQ": 1.0,
        }

        assert unit_config.adsorption_model == "STERIC_MASS_ACTION"
        npt.assert_equal(unit_config.adsorption, expected_adsorption_config)

    def check_particle_config(self, unit_config):
        """
        Check the particle configuration.

        Parameters
        ----------
        unit_config : dict
            The configuration of the unit.
        """
        assert unit_config.PAR_POROSITY == 0.75
        assert unit_config.PAR_RADIUS == 4.5e-05
        assert unit_config.discretization.par_geom == "SPHERE"

    def check_discretization(self, unit, unit_config):
        """
        Check the discretization configuration.

        Parameters
        ----------
        unit : Unit
            The unit object.
        unit_config : dict
            The configuration of the unit.
        """
        assert unit_config.discretization.ncol == unit.discretization.ncol
        assert (
            unit_config.discretization.use_analytic_jacobian
            == unit.discretization.use_analytic_jacobian
        )
        assert unit_config.discretization.reconstruction == "WENO"

        npt.assert_equal(
            unit_config.discretization.weno,
            {"boundary_model": 0, "weno_eps": 1e-10, "weno_order": 3},
        )

        npt.assert_equal(
            unit_config.discretization.consistency_solver,
            {
                "solver_name": "LEVMAR",
                "init_damping": 0.01,
                "min_damping": 0.0001,
                "max_iterations": 50,
                "subsolvers": "LEVMAR",
            },
        )

        if "spatial_method" in unit.discretization.parameters:
            assert (
                unit_config.discretization.spatial_method
                == unit.discretization.spatial_method
            )

        if "npar" in unit.discretization.parameters:
            assert unit_config.discretization.npar == unit.discretization.npar
            assert unit_config.discretization.par_disc_type == "EQUIDISTANT_PAR"

    def test_solver_config(self, process: Process):
        """
        Test the solver configuration for the process.

        Parameters
        ----------
        process : Process
            The process object.
        """
        process_config = self.return_process_config(process)
        solver_config = process_config.solver

        expected_solver_config = {
            "nthreads": 1,
            "consistent_init_mode": 1,
            "consistent_init_mode_sens": 1,
            "user_solution_times": np.arange(0.0, 120 * 60 + 1),
            "sections": {
                "nsec": 3,
                "section_times": [0.0, 10.0, 90.0, 7200.0],
                "section_continuity": [0, 0],
            },
            "time_integrator": {
                "abstol": 1e-08,
                "algtol": 1e-12,
                "reltol": 1e-06,
                "reltol_sens": 1e-12,
                "init_step_size": 1e-06,
                "max_steps": 1000000,
                "max_step_size": 0.0,
                "errortest_sens": False,
                "max_newton_iter": 1000000,
                "max_errtest_fail": 1000000,
                "max_convtest_fail": 1000000,
                "max_newton_iter_sens": 1000000,
            },
        }

        npt.assert_equal(solver_config, expected_solver_config)

    def test_return_config(self, process: Process):
        """
        Test the return configuration for the process.

        Parameters
        ----------
        process : Process
            The process object.
        """
        process_config = self.return_process_config(process)
        return_config = process_config["return"]

        # Assert that all values in return_config.unit_001 (model unit operation) are True
        for key, value in return_config["unit_001"].items():
            assert value, f"The value for key '{key}' is not True. Found: {value}"

    def test_sensitivity_config(self, process: Process):
        """
        Test the sensitivity configuration for the process.

        Parameters
        ----------
        process : Process
            The process object.
        """
        process_config = self.return_process_config(process)
        sensitivity_config = process_config.sensitivity

        expected_sensitivity_config = {"sens_method": "ad1", "nsens": 0}
        npt.assert_equal(sensitivity_config, expected_sensitivity_config)


@pytest.mark.parametrize("simulation_results", simulation_test_cases, indirect=True)
@pytest.mark.slow
class TestResultsWithLWE:
    def test_trigger_simulation(self, simulation_results):
        """
        Test to trigger the simulation.
        """
        simulation_results = simulation_results
        assert simulation_results is not None

    def test_compare_solution_shape(self, simulation_results):
        """
        Compare the dimensions of the solution object against the expected solution shape.
        """
        simulation_results = simulation_results
        process = simulation_results.process
        unit = process.flow_sheet.units[1]

        # for units without ports
        if not unit.has_ports:
            # assert solution inlet has shape (t, n_comp)
            assert simulation_results.solution[unit.name].inlet.solution_shape == (
                int(process.cycle_time + 1),
                process.component_system.n_comp,
            )
            # assert solution outlet has shape (t, n_comp)
            assert simulation_results.solution[unit.name].outlet.solution_shape == (
                int(process.cycle_time + 1),
                process.component_system.n_comp,
            )
            # assert solution bulk has shape (t, n_col, n_comp)
            if not isinstance(unit.discretization, NoDiscretization):
                assert simulation_results.solution[unit.name].bulk.solution_shape == (
                    int(process.cycle_time + 1),
                    unit.discretization.ncol,
                    process.component_system.n_comp,
                )

        # for units with ports
        else:
            # assert solution inlet is given for each port
            assert len(simulation_results.solution[unit.name].inlet) == unit.n_ports
            # assert solution for channel 0 has shape (t, n_comp)
            assert simulation_results.solution[
                unit.name
            ].inlet.channel_0.solution_shape == (
                int(process.cycle_time + 1),
                process.component_system.n_comp,
            )
            # assert solution bulk has shape (t, n_col, n_ports, n_comp)
            assert simulation_results.solution[unit.name].bulk.solution_shape == (
                int(process.cycle_time + 1),
                unit.discretization.ncol,
                unit.n_ports,
                process.component_system.n_comp,
            )

        # for units with particles
        if unit.supports_binding and not isinstance(
            unit.discretization, NoDiscretization
        ):
            # for units with solid phase and particle discretization
            if "npar" in unit.discretization.parameters:
                # assert solution solid has shape (t, n_col, n_par, n_comp)
                assert simulation_results.solution[unit.name].solid.solution_shape == (
                    int(process.cycle_time + 1),
                    unit.discretization.ncol,
                    unit.discretization.npar,
                    process.component_system.n_comp,
                )
            # for units with solid phase and without particle discretization
            else:
                # assert solution solid has shape (t, ncol, n_comp)
                assert simulation_results.solution[unit.name].solid.solution_shape == (
                    int(process.cycle_time + 1),
                    unit.discretization.ncol,
                    process.component_system.n_comp,
                )

        # for units with particle mobile phase and particle discretization
        if (
            unit.supports_particle_reaction
            and unit.name != "LumpedRateModelWithoutPores"
        ):
            # assert soluction particle has shape (t, n_col, n_par, n_comp)
            if "npar" in unit.discretization.parameters:
                assert simulation_results.solution[
                    unit.name
                ].particle.solution_shape == (
                    int(process.cycle_time + 1),
                    unit.discretization.ncol,
                    unit.discretization.npar,
                    process.component_system.n_comp,
                )
            # for units with particle mobiles phase and without particle discretization
            else:
                # assert solution particle has shape (t, n_col, n_comp)
                assert simulation_results.solution[
                    unit.name
                ].particle.solution_shape == (
                    int(process.cycle_time + 1),
                    unit.discretization.ncol,
                    process.component_system.n_comp,
                )


if __name__ == "__main__":
    pytest.main([__file__])
